# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import aacgmv2


def __haversine_distances(target_lat, target_lon, lat_array, lon_array):
    """
    Computes the distance on the globe between target lat/lon,
    and all points defined by lat/lon arrays.
    """
    # Earth radius (meters)
    r = 6371000.0

    # Convert degrees to radians
    phi1 = np.radians(target_lat)
    phi2 = np.radians(lat_array)
    delta_phi = np.radians(lat_array - target_lat)
    delta_lambda = np.radians(lon_array - target_lon)

    # Haversine formula
    a = np.sin(delta_phi / 2.0)**2 + np.cos(phi1) * np.cos(phi2) * np.sin(delta_lambda / 2.0)**2
    c = 2 * np.arctan2(np.sqrt(a), np.sqrt(1 - a))

    return r * c


def mag(skymap, timestamp, altitude_km, contour_lats, contour_lons, constant_lat, constant_lon, n_points, remove_edge_cases):
    # Check that multiple contours are not defined based on arguments passed
    if (contour_lats is None) != (contour_lons is None):
        raise ValueError("When defining a custom contour, both 'contour_lats' and 'contour_lons' must be supplied.")
    if sum([((contour_lats is not None) and (contour_lons is not None)), (constant_lat is not None), (constant_lon is not None)]) > 1:
        raise ValueError("Only one contour can be defined per call: Pass only one of " +
                         "'contour_lats & contour_lons', 'constant_lat', or 'constant_lon'.")
    if sum([((contour_lats is not None) and (contour_lons is not None)), (constant_lat is not None), (constant_lon is not None)]) == 0:
        raise ValueError("No contour defined in input: Pass one of 'contour_lats & contour_lons', 'constant_lat', or 'constant_lon'.")

    # Obtain lat/lon arrays from skymap at desired altitude
    if (altitude_km * 1000.0 in skymap.full_map_altitude):
        altitude_idx = np.where(altitude_km * 1000.0 == skymap.full_map_altitude)

        lats = np.squeeze(skymap.full_map_latitude[altitude_idx, :, :])
        lons = np.squeeze(skymap.full_map_longitude[altitude_idx, :, :])
        lons[np.where(lons > 180)] -= 360.0  # Fix skymap to be in (-180,180) format

    else:
        # Make sure altitude is in range that can be interpolated
        if (altitude_km * 1000.0 < skymap.full_map_altitude[0]) or (altitude_km * 1000.0 > skymap.full_map_altitude[2]):
            raise ValueError("Altitude " + str(altitude_km) + " outside valid range of " +
                             str((skymap.full_map_altitude[0] / 1000.0, skymap.full_map_altitude[2] / 1000.0)))

        # Initialze empty lat/lon arrays
        lats = np.full(np.squeeze(skymap.full_map_latitude[0, :, :]).shape, np.nan, dtype=skymap.full_map_latitude[0, :, :].dtype)
        lons = lats.copy()

        # Interpolate lats and lons at desired altitude
        for i in range(skymap.full_map_latitude.shape[1]):
            for j in range(skymap.full_map_latitude.shape[2]):
                lats[i, j] = np.interp(altitude_km * 1000.0, skymap.full_map_altitude, skymap.full_map_latitude[:, i, j])
                lons[i, j] = np.interp(altitude_km * 1000.0, skymap.full_map_altitude, skymap.full_map_longitude[:, i, j])

        lons[np.where(lons > 180)] -= 360.0  # Fix skymap to be in (-180,180) format

        # Convert to magnetic coords
        mag_lats, mag_lons, _ = aacgmv2.convert_latlon_arr(
            lats.flatten(),
            lons.flatten(),
            (lons * 0.0).flatten(),
            timestamp,
            method_code='G2A',
        )
        lats = np.reshape(mag_lats, lats.shape)
        lons = np.reshape(mag_lons, lons.shape)

    # First handle case of a contour of constant latitude:
    if (constant_lat is not None):
        # First check that the supplied latitude is valid for this skymap
        if (constant_lat < np.nanmin(lats)) or (constant_lat > np.nanmax(lats)):  # pragma: nocover
            raise ValueError(f"Latitude {constant_lat} does not coincide with input skymap: magnetic latitude " +
                             f"range at {altitude_km} km is {np.nanmin(lats), np.nanmax(lats)}.")

        # Get the longitude bounds of the skymap
        min_skymap_lon, max_skymap_lon = np.nanmin(lons), np.nanmax(lons)

        # Determine the step in longitudes for grabbing target lat, based on n_points
        if (n_points is None):
            n_points = round((lats.shape[0] - 1) / 5.12)

        # Iterate though longitude ranges
        iteration_lons = np.linspace(min_skymap_lon, max_skymap_lon, n_points)
        x_list = []
        y_list = []
        for i in range(len(iteration_lons) - 1):
            # Get lon domain for current iteration through skymap
            lon_min = iteration_lons[i]
            lon_max = iteration_lons[i + 1]

            # Get indices of this longitude slice
            lon_slice_idx = np.logical_and(lons >= lon_min, lons < lon_max)

            # Get index of pixel nearest to target elevation
            masked_lats = np.where(lon_slice_idx, lats, np.nan)
            diffs = np.abs(masked_lats - constant_lat)
            y, x = np.where(diffs == np.nanmin(diffs))

            if x.shape == (0, ) or y.shape == (0, ):  # pragma: nocover
                continue

            # Add to master lists
            x_list.append(x[0])
            y_list.append(y[0])

        if (remove_edge_cases is True):
            # Remove any points lying on the edge of CCD bounds and return
            x_list = np.array(x_list)
            y_list = np.array(y_list)
            edge_case_idx = np.where(np.logical_and.reduce([x_list > 0, x_list < lats.shape[1] - 1, y_list > 0, y_list < lats.shape[0] - 1]))
            x_list = x_list[edge_case_idx]
            y_list = y_list[edge_case_idx]
            return (x_list, y_list)
        else:
            # Convert to arrays, return
            return (np.array(x_list), np.array(y_list))

    # Next handle case of a contour of constant longitude:
    elif (constant_lon is not None):
        # First check that the supplied longitude is valid for this skymap
        if (constant_lon < np.nanmin(lons)) or (constant_lon > np.nanmax(lons)):  # pragma: nocover
            raise ValueError(f"Longitude {constant_lat} does not coincide with input skymap: magnetic longitude " +
                             f"range at {altitude_km} km is {np.nanmin(lons), np.nanmax(lons)}.")

        # Get the latitude bounds of the skymap
        min_skymap_lat, max_skymap_lat = np.nanmin(lats), np.nanmax(lats)

        # Determine the step in longitudes for grabbing target lat, based on n_points
        if (n_points is None):
            n_points = round((lats.shape[0] - 1) / 5.12)

        # Iterate though latitude ranges
        iteration_lats = np.linspace(min_skymap_lat, max_skymap_lat, n_points)
        x_list = []
        y_list = []
        for i in range(len(iteration_lats) - 1):
            # Get lon domain for current iteration through skymap
            lat_min = iteration_lats[i]
            lat_max = iteration_lats[i + 1]

            # Get indices of this longitude slice
            lat_slice_idx = np.logical_and(lats >= lat_min, lats < lat_max)

            # Get index of pixel nearest to target elevation
            masked_lons = np.where(lat_slice_idx, lons, np.nan)
            diffs = np.abs(masked_lons - constant_lon)
            y, x = np.where(diffs == np.nanmin(diffs))

            if x.shape == (0, ) or y.shape == (0, ):  # pragma: nocover
                continue

            # Add to master lists
            x_list.append(x[0])
            y_list.append(y[0])

        if (remove_edge_cases is True):
            # Remove any points lying on the edge of CCD bounds and return
            x_list = np.array(x_list)
            y_list = np.array(y_list)
            edge_case_idx = np.where(np.logical_and.reduce([x_list > 0, x_list < lats.shape[1] - 1, y_list > 0, y_list < lats.shape[0] - 1]))
            x_list = x_list[edge_case_idx]
            y_list = y_list[edge_case_idx]
            return (x_list, y_list)
        else:
            # Convert to arrays, return
            return (np.array(x_list), np.array(y_list))

    # Finally, handle case of a custom contour
    elif (contour_lats is not None) and (contour_lons is not None):
        # Convert lists to ndarrays if necessary
        if (isinstance(contour_lats, list)):
            contour_lats = np.asarray(contour_lats)
        if (isinstance(contour_lons, list)):
            contour_lons = np.asarray(contour_lons)

        # Remove any invalid lat/lon pairs
        invalid_mask = (contour_lons < np.nanmin(lons)) | (contour_lons > np.nanmax(lons)) | (contour_lats < np.nanmin(lats)) | (contour_lats
                                                                                                                                 > np.nanmax(lats))
        # Filter out invalid contour points
        valid_contour_lats = contour_lats[~invalid_mask]
        valid_contour_lons = contour_lons[~invalid_mask]

        # Check if any lat/lons are actually valid for skymap
        invalid_lats_mask = (contour_lats < np.nanmin(lats)) | (contour_lats > np.nanmax(lats))
        invalid_lons_mask = (contour_lons < np.nanmin(lons)) | (contour_lons > np.nanmax(lons))
        if contour_lats[~invalid_lats_mask].shape == (0, ):  # pragma: nocover
            raise ValueError(f"Magnetic latitudes provided are outside this skymap's valid range of {(np.nanmin(lats),np.nanmax(lats))}.")
        if contour_lons[~invalid_lons_mask].shape == (0, ):  # pragma: nocover
            raise ValueError(f"Magnetic longitudes provided are outside this skymap's valid range of {(np.nanmin(lons),np.nanmax(lons))}.")

        # Get the latitude bounds of the skymap
        # Iterate through each target point
        x_list = []
        y_list = []
        for target_lat, target_lon in zip(valid_contour_lats, valid_contour_lons):
            # Make sure lat/lon falls within skymap
            if target_lat < np.nanmin(lats) or target_lat > np.nanmax(lats):  # pragma: nocover
                raise ValueError(f"Magnetic latitude {target_lat} is outside this skymap's valid range of {(np.nanmin(lats),np.nanmax(lats))}.")
            if target_lon < np.nanmin(lons) or target_lon > np.nanmax(lons):  # pragma: nocover
                raise ValueError(f"Magnetic longitude {target_lon} is outside this skymap's valid range of {(np.nanmin(lons),np.nanmax(lons))}.")

            # Compute haversine distance between all points in skymap
            haversine_diff = __haversine_distances(target_lat, target_lon, lats, lons)

            # Obtain the skymap indices of the nearest point
            nan_mask = np.isnan(haversine_diff)
            masked_data = np.ma.masked_array(haversine_diff, mask=nan_mask)
            nearest_indices = np.unravel_index(np.ma.argmin(masked_data), haversine_diff.shape)

            # Convert indices to CCD Coordinates
            y_loc = nearest_indices[0] - 1
            if y_loc < 0:  # pragma: nocover
                y_loc = 0
            x_loc = nearest_indices[1] - 1
            if x_loc < 0:  # pragma: nocover
                x_loc = 0

            x_list.append(x_loc)
            y_list.append(y_loc)

        if (remove_edge_cases is True):
            # Remove any points lying on the edge of CCD bounds and return
            x_list = np.array(x_list)
            y_list = np.array(y_list)
            edge_case_idx = np.where(np.logical_and.reduce([x_list > 0, x_list < lats.shape[1] - 1, y_list > 0, y_list < lats.shape[0] - 1]))
            x_list = x_list[edge_case_idx]
            y_list = y_list[edge_case_idx]
            return (x_list, y_list)
        else:
            # Convert to arrays, return
            return (np.array(x_list), np.array(y_list))

    else:  # pragma: nocover
        # This shouldn't occur, but typing claims there is a missed case somewhere that could not be identified...
        raise ValueError("Something unexpected happened, please verify your inputs are in expected format. Otherwise, contact the PyAuroraX team.")



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import matplotlib as mpl
import matplotlib.pyplot as plt


def set_theme(theme):
    if (theme == "default"):
        mpl.rcParams.update(mpl.rcParamsDefault)
    elif (theme == "light"):
        plt.style.use("default")
    elif (theme == "dark"):
        plt.style.use("dark_background")
    else:
        plt.style.use(theme)



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unique exception classes utilized by PyAuroraX. These exceptions can be used to help trap specific 
errors raised by this library.

Note that all exceptions are imported at the root level of the library. They
can be referenced using [`pyaurorax.AuroraXError`](exceptions.html#pyaurorax.exceptions.AuroraXError) 
or `pyaurorax.exceptions.AuroraXError`.
"""


class AuroraXError(Exception):

    def __init__(self, *args, **kwargs):
        super(AuroraXError, self).__init__(*args, **kwargs)  # pragma: no cover


class AuroraXInitializationError(AuroraXError):
    """
    Error occurred during library initialization
    """
    pass


class AuroraXPurgeError(AuroraXError):
    """
    Error occurred during purging of download or tar extraction working directory
    """
    pass


class AuroraXAPIError(AuroraXError):
    """
    Error occurred during an API call
    """
    pass


class AuroraXNotFoundError(AuroraXError):
    """
    The AuroraX record was not found
    """
    pass


class AuroraXDuplicateError(AuroraXError):
    """
    A duplicate record already exists
    """
    pass


class AuroraXUnauthorizedError(AuroraXError):
    """
    A privileged operation was attempted without authorization
    """
    pass


class AuroraXConflictError(AuroraXError):
    """
    A conflict occurred while modifying records
    """
    pass


class AuroraXDataRetrievalError(AuroraXError):
    """
    Error occurred while retrieving search data
    """
    pass


class AuroraXSearchError(AuroraXError):
    """
    An error occurred in the API while performing a search
    """
    pass


class AuroraXUploadError(AuroraXError):
    """
    Error occurred during upload operation
    """
    pass


class AuroraXMaintenanceError(AuroraXError):
    """
    AuroraX API is in maintenance mode, read-only tasks are only possible
    """
    pass


class AuroraXUnsupportedReadError(AuroraXError):
    """
    Unsupported dataset for read function

    NOTE: this is primarily a PyUCalgarySRS error
    """
    pass


class AuroraXDownloadError(AuroraXError):
    """
    Error occurred during downloading of data

    NOTE: this is primarily a PyUCalgarySRS error
    """
    pass



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import aacgmv2
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.path import Path
from ..classes.keogram import Keogram
from ...data.ucalgary import Skymap


# Helper function that returns all array indices within
# the polygon defined by input vertices.
def __indices_in_polygon(vertices, image_shape):
    """
    Function to obtain all indices of an array/image, within the 
    polygon defined by input list of ordered vertices.
    """
    # Create a Path object using the vertices
    path = Path(vertices)

    # Create a grid of points covering the image shape
    x, y = np.meshgrid(np.arange(image_shape[1]), np.arange(image_shape[0]))
    points = np.vstack((x.flatten(), y.flatten())).T

    # Create a mask indicating which points are inside the polygon
    mask = path.contains_points(points).reshape(image_shape)

    # Get all indices inside or on the boundary of the polygon
    indices_inside = np.argwhere(mask)

    return indices_inside


def __haversine_distances(target_lat, target_lon, lat_array, lon_array):
    """
    Computes the distance on the globe between target lat/lon,
    and all points defined by lat/lon arrays.
    """
    # Earth radius (meters)
    r = 6371000.0

    # Convert degrees to radians
    phi1 = np.radians(target_lat)
    phi2 = np.radians(lat_array)
    delta_phi = np.radians(lat_array - target_lat)
    delta_lambda = np.radians(lon_array - target_lon)

    # Haversine formula
    a = np.sin(delta_phi / 2.0)**2 + np.cos(phi1) * np.cos(phi2) * np.sin(delta_lambda / 2.0)**2
    c = 2 * np.arctan2(np.sqrt(a), np.sqrt(1 - a))

    return r * c


# Helper function for handling magnetic lat/lon inputs
def __convert_latlon_to_ccd(lon_locs, lat_locs, timestamp, skymap: Skymap, altitude_km, magnetic):
    """
    Function for handling lat/lon inputs. Takes the lat/lon
    locations, and uses a skymap to convert those locations
    to CCD (image index) coordinates.
    """
    # Obtain lat/lon arrays from skymap
    if (altitude_km * 1000.0 in skymap.full_map_altitude):
        altitude_idx = np.where(altitude_km * 1000.0 == skymap.full_map_altitude)

        lats = np.squeeze(skymap.full_map_latitude[altitude_idx, :, :])
        lons = np.squeeze(skymap.full_map_longitude[altitude_idx, :, :])
        lons[np.where(lons > 180)] -= 360.0

    else:
        # Make sure altitude is in range that can be interpolated
        if (altitude_km * 1000.0 < skymap.full_map_altitude[0]) or (altitude_km * 1000.0 > skymap.full_map_altitude[2]):
            raise ValueError("Altitude " + str(altitude_km) + " outside valid range of " +
                             str((skymap.full_map_altitude[0] / 1000.0, skymap.full_map_altitude[2] / 1000.0)))

        # Initialize empty lat/lon arrays
        lats = np.full(np.squeeze(skymap.full_map_latitude[0, :, :]).shape, np.nan, dtype=skymap.full_map_latitude[0, :, :].dtype)
        lons = lats.copy()

        # Interpolate lats and lons at desired altitude
        for i in range(skymap.full_map_latitude.shape[1]):
            for j in range(skymap.full_map_latitude.shape[2]):
                lats[i, j] = np.interp(altitude_km * 1000.0, skymap.full_map_altitude, skymap.full_map_latitude[:, i, j])
                lons[i, j] = np.interp(altitude_km * 1000.0, skymap.full_map_altitude, skymap.full_map_longitude[:, i, j])

        lons[np.where(lons > 180)] -= 360.0  # Fix skymap to be in (-180,180) format

    # Convert skymap to magnetic coords if necessary
    if magnetic:
        mag_lats, mag_lons, _ = aacgmv2.convert_latlon_arr(
            lats.flatten(),
            lons.flatten(),
            (lons * 0.0).flatten(),
            timestamp[0],
            method_code="G2A",
        )
        lats = np.reshape(mag_lats, lats.shape)
        lons = np.reshape(mag_lons, lons.shape)

    # Iterate through each target point
    x_locs = []
    y_locs = []
    for target_lat, target_lon in zip(lat_locs, lon_locs):

        # Make sure lat/lon falls within skymap
        if target_lat < np.nanmin(lats) or target_lat > np.nanmax(lats):
            continue
            # raise ValueError(f"Latitude {target_lat} is outside this skymap's valid range of {(np.nanmin(lats),np.nanmax(lats))}.")
        if target_lon < np.nanmin(lons) or target_lon > np.nanmax(lons):
            continue
            # raise ValueError(f"Longitude {target_lon} is outside this skymap's valid range of {(np.nanmin(lons),np.nanmax(lons))}.")

        # Compute haversine distance between all points in skymap
        haversine_diff = __haversine_distances(target_lat, target_lon, lats, lons)

        # Obtain the skymap indices of the nearest point
        nan_mask = np.isnan(haversine_diff)
        masked_data = np.ma.masked_array(haversine_diff, mask=nan_mask)
        nearest_indices = np.unravel_index(np.ma.argmin(masked_data), haversine_diff.shape)

        # Convert indices to CCD Coordinates
        y_loc = nearest_indices[0] - 1
        if y_loc < 0:
            y_loc = 0
        x_loc = nearest_indices[1] - 1
        if x_loc < 0:
            x_loc = 0

        x_locs.append(x_loc)
        y_locs.append(y_loc)

    x_locs = np.array(x_locs)
    y_locs = np.array(y_locs)

    return x_locs, y_locs


def create_custom(images, timestamp, coordinate_system, width, x_locs, y_locs, preview, skymap, altitude_km, metric):
    # If using CCD coordinates we don't need a skymao or altitude
    if (coordinate_system == "ccd") and (skymap is not None or altitude_km is not None):
        raise ValueError("Conflict in passing in a skymap or altitude when working in CCD coordinates. This value will be ignored.")

    # convert any lists to np.arrays  and check shape
    x_locs = np.array(x_locs)
    y_locs = np.array(y_locs)

    # determine if we are single or 3 channel
    n_channels = 1
    if (len(images.shape) == 3):
        # single channel
        n_channels = 1
    elif (len(images.shape) == 4):
        # three channel
        n_channels = 3
    else:
        ValueError("Unable to determine number of channels based on the supplied images. Make sure you are supplying a " +
                   "[rows,cols,images] or [rows,cols,channels,images] sized array.")

    # Initialize empty keogram array
    keo_arr = np.squeeze(np.full((x_locs.shape[0] - 1, len(timestamp), n_channels), 0))

    if len(x_locs.shape) != 1:
        raise ValueError(f"X coordinates may not be multidimensional. Sequence passed with shape {x_locs.shape}")
    if len(y_locs.shape) != 1:
        raise ValueError(f"Y coordinates may not be multidimensional. Sequence passed with shape {y_locs.shape}")
    if len(x_locs.shape) != len(y_locs.shape):
        raise ValueError(f"X and Y coordinates must have same length. Sequences passed with shapes {x_locs.shape} and {y_locs.shape}")

    # Convert lat/lon coordinates to CCD
    if coordinate_system == "mag":
        if (skymap is None or altitude_km is None):
            raise ValueError("When magnetic coordinates, a Skymap object and Altitude must be passed in through the skymap argument.")
        x_locs, y_locs = __convert_latlon_to_ccd(x_locs, y_locs, timestamp, skymap, altitude_km, magnetic=True)
    elif coordinate_system == "geo":
        if (skymap is None or altitude_km is None):
            raise ValueError("When geographic coordinates, a Skymap object and Altitude must be passed in through the skymap argument.")
        x_locs, y_locs = __convert_latlon_to_ccd(x_locs, y_locs, timestamp, skymap, altitude_km, magnetic=False)

    # We will use the first image as a preview
    if n_channels == 1:
        preview_img = images.copy()[:, :, 0]
    else:
        preview_img = images.copy()[:, :, :, 0]

    # Now working in CCD Coordinates
    x_max = images.shape[1] - 1
    y_max = images.shape[0] - 1

    # Remove any points that are not within the image CCD
    parsed_x_locs = []
    parsed_y_locs = []
    for i in range(x_locs.shape[0]):
        x = x_locs[i]
        y = y_locs[i]

        if x < 0 or x > x_max:
            continue
        if y < 0 or y > y_max:
            continue

        parsed_x_locs.append(x)
        parsed_y_locs.append(y)
    x_locs = np.array(parsed_x_locs)
    y_locs = np.array(parsed_y_locs)

    # Make sure all supplied points are within image bounds # NOTE: This should be good to remove as it shouldn't hit, testing needed
    if len(np.where(np.logical_or(x_locs < 0, x_locs > x_max))[0]) != 0:
        raise ValueError("The following CCD coordinates passed in through x_locs are outside of the CCD image range: " +
                         str(x_locs[np.where(np.logical_or(x_locs < 0, x_locs > x_max))]))
    if len(np.where(np.logical_or(y_locs < 0, y_locs > y_max))[0]) != 0:
        raise ValueError("The following CCD coordinates passed in through y_locs are outside of the CCD image range: " +
                         str(y_locs[np.where(np.logical_or(y_locs < 0, y_locs > y_max))]))

    # Iterate points in pairs of two
    path_counter = 0
    for i in range(x_locs.shape[0] - 1):

        # Points of concern for this iteration
        x_0 = x_locs[i]
        x_1 = x_locs[i + 1]
        y_0 = y_locs[i]
        y_1 = y_locs[i + 1]

        # Compute the unit vector between the two points
        dx = x_1 - x_0
        dy = y_1 - y_0
        length = np.sqrt(dx**2 + dy**2)
        if length == 0:
            continue

        dx /= length
        dy /= length

        # Compute orthogonal unit vector
        perp_dx = -dy
        perp_dy = dx

        # Calculate (+/-) offsets for each perpendicular direction
        offset1_x = perp_dx * width / 2
        offset1_y = perp_dy * width / 2
        offset2_x = -perp_dx * width / 2
        offset2_y = -perp_dy * width / 2

        # Calculate vertices in correct order for this polygon
        vertex1 = (int(x_0 + offset1_x), int(y_0 + offset1_y))
        vertex2 = (int(x_1 + offset1_x), int(y_1 + offset1_y))
        vertex3 = (int(x_1 + offset2_x), int(y_1 + offset2_y))
        vertex4 = (int(x_0 + offset2_x), int(y_0 + offset2_y))

        # Append vertices in the correct order to form a closed polygon
        vertices = [vertex1, vertex2, vertex3, vertex4]

        # Obtain the indexes into the image of this polygon
        indices_inside = __indices_in_polygon(vertices, (images.shape[0], images.shape[1]))

        if np.any(np.array(indices_inside.shape) == 0):
            continue

        row_idx, col_idx = zip(*indices_inside)

        if n_channels == 1:
            # Update the preview image
            preview_img[row_idx, col_idx] = np.iinfo(preview_img.dtype).max

            # Extract metric from all images and add to keogram
            if metric == "median":
                # median metric
                pixel_keogram = np.median(images[row_idx, col_idx, :], axis=0)
            elif metric == "mean":
                # mean metric
                pixel_keogram = np.mean(images[row_idx, col_idx, :], axis=0)
            else:
                # sum metric
                pixel_keogram = np.sum(images[row_idx, col_idx, :], axis=0)
            keo_arr[i, :] = pixel_keogram
        elif n_channels == 3:
            # Update the preview image
            preview_img[row_idx, col_idx, :] = np.iinfo(preview_img.dtype).max

            # Extract metric from all images and add to keogram
            if metric == "median":
                r_pixel_keogram = np.floor(np.median(images[row_idx, col_idx, 0, :], axis=0))
                g_pixel_keogram = np.floor(np.median(images[row_idx, col_idx, 1, :], axis=0))
                b_pixel_keogram = np.floor(np.median(images[row_idx, col_idx, 2, :], axis=0))
            elif metric == "mean":
                r_pixel_keogram = np.floor(np.mean(images[row_idx, col_idx, 0, :], axis=0))
                g_pixel_keogram = np.floor(np.mean(images[row_idx, col_idx, 1, :], axis=0))
                b_pixel_keogram = np.floor(np.mean(images[row_idx, col_idx, 2, :], axis=0))
            else:
                # sum metric
                r_pixel_keogram = np.floor(np.sum(images[row_idx, col_idx, 0, :], axis=0))
                g_pixel_keogram = np.floor(np.sum(images[row_idx, col_idx, 1, :], axis=0))
                b_pixel_keogram = np.floor(np.sum(images[row_idx, col_idx, 2, :], axis=0))

            keo_arr[i, :, 0] = r_pixel_keogram
            keo_arr[i, :, 1] = g_pixel_keogram
            keo_arr[i, :, 2] = b_pixel_keogram

        path_counter += 1

    if path_counter == 0:
        raise ValueError("Could not form keogram path. First ensure that coordinates are within image range. Then " +
                         "try increasing 'width' or decreasing number of points in input coordinates.")

    # Create keogram object
    keo_obj = Keogram(data=keo_arr, timestamp=timestamp, instrument_type="asi")

    # show preview
    if preview:
        plt.figure()
        plt.imshow(preview_img, cmap="gray", origin="lower")
        plt.axis("off")
        plt.title("Keogram Domain Preview")
        plt.show()

    # return
    return keo_obj



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Perform various calibration procedures on image data. 
"""

import numpy as np
from typing import Optional
from ...data.ucalgary import Calibration
from ._rego import apply_calibration as func_rego
from ._trex_nir import apply_calibration as func_trex_nir

__all__ = ["CalibrationManager"]


class CalibrationManager:
    """
    The CalibrationManager object is initialized within every PyAuroraX object. It acts as a way to access 
    the submodules and carry over configuration information in the super class.
    """

    def __init__(self):
        pass

    def rego(
        self,
        images: np.ndarray,
        cal_flatfield: Optional[Calibration] = None,
        cal_rayleighs: Optional[Calibration] = None,
        step_dark_frame_correction: bool = True,
        step_flatfield_calibration: bool = True,
        step_rayleighs_calibration: bool = True,
        exposure_length_sec: float = 2.0,
    ) -> np.ndarray:
        """
        Apply various calibration adjustments to a single or set of images raw images.

        Args:
            images (numpy.ndarray): 
                Raw images to perform calibration procedures on.
            cal_flatfield (pyaurorax.data.ucalgary.Calibration): 
                Calibration object containing the flatfield data to utilize. This field is required if
                the `step_flatfield_corection` is set to True.
            cal_rayleighs (pyaurorax.data.ucalgary.Calibration): 
                Calibration object containing the Rayleighs data to utilize. This field is required if 
                the `step_rayleighs_calibration` is set to True.
            step_dark_frame_correction (bool): 
                Perform the dark frame correction step by subtracting an average of a bottom corner grid from 
                the image (ie. 4x4.). Defaults to `True`.
            step_flatfield_calibration (bool): 
                Perform the flatfield correction step. Defaults to `True`. Note that the `cal_flatfield` parameter
                must be supplied if this is True.
            step_rayleighs_calibration (bool): 
                Perform the Rayleighs conversion step. Defaults to `True.` Note that the `cal_rayleighs` parameter
                must be supplied if this is True.
            exposure_length_sec (float): 
                Force the exposure length to be a certain value. Default is TREx NIR's nominal operating mode 
                exposure length of `2.0 seconds`. Adjusting this field should be done with caution.

        Returns:
            The calibrated images. 
            
            The shape of the calibrated data will be same as the input images. The dtype of the calibrated data 
            will depend on if the Rayleighs conversion was performed. If it was, a float32 array will be returned. 
            If it wasn't, the dtype will be the same as input images' dtype.

        Raises:
            ValueError: issues encountered with supplied parameters.
        """
        return func_rego(
            images,
            cal_flatfield,
            cal_rayleighs,
            step_dark_frame_correction,
            step_flatfield_calibration,
            step_rayleighs_calibration,
            exposure_length_sec,
        )

    def trex_nir(
        self,
        images: np.ndarray,
        cal_flatfield: Optional[Calibration] = None,
        cal_rayleighs: Optional[Calibration] = None,
        step_dark_frame_correction: bool = True,
        step_flatfield_calibration: bool = True,
        step_rayleighs_calibration: bool = True,
        exposure_length_sec: float = 5.0,
    ) -> np.ndarray:
        """
        Apply various calibration adjustments to a single or set of images raw images.

        Args:
            images (numpy.ndarray): 
                Raw images to perform calibration procedures on.
            cal_flatfield (pyaurorax.data.ucalgary.Calibration): 
                Calibration object containing the flatfield data to utilize. This field is required if
                the `step_flatfield_corection` is set to True.
            cal_rayleighs (pyaurorax.data.ucalgary.Calibration): 
                Calibration object containing the Rayleighs data to utilize. This field is required if 
                the `step_rayleighs_calibration` is set to True.
            step_dark_frame_correction (bool): 
                Perform the dark frame correction step. Defaults to `True`.
            step_flatfield_calibration (bool): 
                Perform the flatfield correction step. Defaults to `True`. Note that the `cal_flatfield` parameter
                must be supplied if this is True.
            step_rayleighs_calibration (bool): 
                Perform the Rayleighs conversion step. Defaults to `True.` Note that the `cal_rayleighs` parameter
                must be supplied if this is True.
            exposure_length_sec (float): 
                Force the exposure length to be a certain value. Default is TREx NIR's nominal operating mode 
                exposure length of `5.0 seconds`. Adjusting this field should be done with caution.

        Returns:
            The calibrated images. 
            
            The shape of the calibrated data will be same as the input images. The dtype of the calibrated data 
            will depend on if the Rayleighs conversion was performed. If it was, a float32 array will be returned. 
            If it wasn't, the dtype will be the same as input images' dtype.

        Raises:
            ValueError: issues encountered with supplied parameters.
        """
        return func_trex_nir(
            images,
            cal_flatfield,
            cal_rayleighs,
            step_dark_frame_correction,
            step_flatfield_calibration,
            step_rayleighs_calibration,
            exposure_length_sec,
        )



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Class definition for a metadata filter objects
"""

from typing import List, Literal, Union, Any


class MetadataFilterExpression:
    """
    Representation for an AuroraX search engine metadata filter expression. These are used 
    as part of conjunction, ephemeris, and data product searching.

    Attributes:
        key (str): 
            The special key for the metadata filter. For example, 'nbtrace_region'.
        
        values (Any or List[Any]): 
            The value(s) that the search will use when filtering. This can either be a single value, 
            or a list of values.

        operator (str): 
            The operator to use when the search engine evaluates the expression. Valid choices
            are: "=", "!=", ">", "<", ">=", "<=", "between", "in", "not in".

            The "in" and "not in" operators are meant exclusively for expressions where there 
            are multiple values (ie. the values parameter is a list of strings).
        
    Raises:
        ValueError: if invalid operator was specified.
    """

    def __init__(
        self,
        key: str,
        values: Union[Any, List[Any]],
        operator: Literal["=", "!=", ">", "<", ">=", "<=", "between", "in", "not in"] = "in",
    ):
        # set required parameters
        self.key = key
        self.values = values

        # set operator
        if (operator not in ["=", "!=", ">", "<", ">=", "<=", "between", "in", "not in"]):
            raise ValueError(
                "Operator '%s' not allowed. You must use one of the following: ['=', '!=', '>', '<', '>=', '<=', 'between', 'in', 'not in']" %
                (operator))
        self.__operator = operator

    @property
    def operator(self) -> str:
        return self.__operator

    @operator.setter
    def operator(self, val: Literal["=", "!=", ">", "<", ">=", "<=", "between", "in", "not in"]):
        if (val not in ["=", "!=", ">", "<", ">=", "<=", "between", "in", "not in"]):
            raise ValueError(
                "Operator '%s' not allowed. You must use one of the following: ['=', '!=', '>', '<', '>=', '<=', 'between', 'in', 'not in']" % (val))
        self.__operator = val

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        # set special strings
        values_str = "'%s'" % (self.values)
        if (type(self.values) is list):
            values_str = str(self.values)

        # return
        return "MetadataFilterExpression(key='%s', values=%s, operator='%s')" % (self.key, values_str, self.operator)

    def pretty_print(self):
        """
        A special print output for this class.
        """

        # print
        print("MetadataFilterExpression:")
        print("  %-10s: %s" % ("key", self.key))
        print("  %-10s: %s" % ("values", self.values))
        print("  %-10s: %s" % ("operator", self.operator))

    def to_query_dict(self):
        """
        Convert the expression object to a dictionary that will be used when executing a search.
        """
        return {
            "key": str(self.key),
            "values": [str(self.values)] if type(self.values) is not list else self.values,
            "operator": str(self.operator),
        }


class MetadataFilter:
    """
    Representation for an AuroraX search engine metadata filter. These are used 
    as part of conjunction, ephemeris, and data product searching.

    Attributes:
        expressions (List[MetadataFilterExpression]): 
            The list of metadata filter expressions for use with conjunction, ephemeris, and 
            data product searches.
        
        operator (str): 
            The logical operator to use when the search engine will evaluate multiple expressions. If 
            not supplied, the search engine will perform a logical 'AND' between each expression. Possible
            choices are 'and' or 'or'.

    Raises:
        ValueError: if invalid operator was specified.
    """

    def __init__(
        self,
        expressions: List[MetadataFilterExpression],
        operator: Literal["and", "or", "AND", "OR"] = "and",
    ):
        self.expressions = expressions

        # set operator
        if (operator.lower() not in ["and", "or"]):
            raise ValueError("Operator '%s' not allowed. You must use one of the following: ['and', 'or']" % (operator))
        self.__operator = operator

    @property
    def operator(self):
        return self.__operator

    @operator.setter
    def operator(self, val: Literal["and", "or", "AND", "OR"] = "and"):
        if (val.lower() not in ["and", "or"]):
            raise ValueError("Operator '%s' not allowed. You must use one of the following: ['and', 'or']" % (val))
        self.__operator = val.lower()

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        # set special strings
        if (len(self.expressions) == 1):
            expressions_str = "[1 expression]"
        else:
            expressions_str = "[%d expressions]" % (len(self.expressions))

        # return
        return "MetadataFilter(expressions=%s, operator='%s')" % (expressions_str, self.operator)

    def pretty_print(self):
        """
        A special print output for this class.
        """
        # set special strings
        if (len(self.expressions) == 1):
            expressions_str = "[1 expression]"
        else:
            expressions_str = "[%d expressions]" % (len(self.expressions))

        # print
        print("MetadataFilter:")
        print("  %-13s: %s" % ("expressions", expressions_str))
        print("  %-13s: %s" % ("operator", self.operator))

    def to_query_dict(self):
        """
        Convert the expression object to a dictionary that will be used when executing a search.
        """
        return {
            "expressions": [x.to_query_dict() for x in self.expressions],
            "logical_operator": self.operator.upper(),
        }



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Data analysis toolkit for working with all-sky imager data available within the
AuroraX platform.

This portion of the PyAuroraX library allows you to easily generate basic plots
for ASI data, and common manipulations. These include things like displaying single
images, making keograms, projecting ASI data onto maps, and extracting metrics for
a given lat/lon bounding box.

Example:
    For shorter function calls, you can initialize the tools submodule using like so:

    ```
    import pyaurorax
    aurorax = pyaurorax.PyAuroraX()
    at = aurorax.tools
    ```
"""

# pull in classes
from .classes.keogram import Keogram
from .classes.montage import Montage
from .classes.mosaic import Mosaic, MosaicData, MosaicSkymap

# imports for this file
from .bounding_box import BoundingBoxManager
from .calibration import CalibrationManager
from .ccd_contour import CCDContourManager
from .grid_files import GridFilesManager
from .keogram import KeogramManager
from .montage import MontageManager
from .mosaic import MosaicManager
from .spectra import SpectraManager
from ._display import display as func_display
from ._movie import movie as func_movie
from ._scale_intensity import scale_intensity as func_scale_intensity
from ._util import set_theme as func_set_theme

# typing imports
import numpy as np
from typing import Literal, Optional, Tuple, Union, Any, List

__all__ = [
    "ToolsManager",
    "Keogram",
    "Montage",
    "Mosaic",
    "MosaicData",
    "MosaicSkymap",
]


class ToolsManager:
    """
    The ToolsManager object is initialized within every PyAuroraX object. It acts as a way to access 
    the submodules and carry over configuration information in the super class.
    """

    def __init__(self, aurorax_obj):
        self.__aurorax_obj = aurorax_obj

        # initialize sub-modules
        self.__bounding_box = BoundingBoxManager(self.__aurorax_obj)
        self.__calibration = CalibrationManager()
        self.__ccd_contour = CCDContourManager()
        self.__grid_files = GridFilesManager(self.__aurorax_obj)
        self.__keogram = KeogramManager()
        self.__montage = MontageManager()
        self.__mosaic = MosaicManager(self.__aurorax_obj)
        self.__spectra = SpectraManager()

    # ------------------------------------------
    # properties for submodule managers
    # ------------------------------------------
    @property
    def bounding_box(self):
        """
        Access to the `bounding_box` submodule from within a PyAuroraX object.
        """
        return self.__bounding_box

    @property
    def calibration(self):
        """
        Access to the `calibration` submodule from within a PyAuroraX object.
        """
        return self.__calibration

    @property
    def ccd_contour(self):
        """
        Access to the `ccd_contour` submodule from within a PyAuroraX object.
        """
        return self.__ccd_contour

    @property
    def grid_files(self):
        """
        Access to the `grid_files` submodule from within a PyAuroraX object.
        """
        return self.__grid_files

    @property
    def keogram(self):
        """
        Access to the `keogram` submodule from within a PyAuroraX object.
        """
        return self.__keogram

    @property
    def montage(self):
        """
        Access to the `montage` submodule from within a PyAuroraX object.
        """
        return self.__montage

    @property
    def mosaic(self):
        """
        Access to the `mosaic` submodule from within a PyAuroraX object.
        """
        return self.__mosaic

    @property
    def spectra(self):
        """
        Access to the `spectra` submodule from within a PyAuroraX object.
        """
        return self.__spectra

    # ------------------------------------------
    # functions available at this manager level
    # ------------------------------------------
    def display(self,
                image: np.ndarray,
                cmap: Optional[str] = None,
                figsize: Optional[Tuple[int, int]] = None,
                aspect: Optional[Union[Literal["equal", "auto"], float]] = None,
                colorbar: bool = False,
                title: Optional[str] = None,
                returnfig: bool = False,
                savefig: bool = False,
                savefig_filename: Optional[str] = None,
                savefig_quality: Optional[int] = None) -> Any:
        """
        Render a visualization of a single image.

        Either display it (default behaviour), save it to disk (using the `savefig` parameter), or 
        return the matplotlib plot object for further usage (using the `returnfig` parameter).

        Args:
            image (numpy.ndarray): 
                The image to display, represented as a numpy array.

            cmap (str): 
                The matplotlib colormap to use.

                Commonly used colormaps are:

                - REGO: `gist_heat`
                - THEMIS ASI: `gray`
                - TREx Blue: `Blues_r`
                - TREx NIR: `gray`
                - TREx RGB: `None`

                A list of all available colormaps can be found on the 
                [matplotlib documentation](https://matplotlib.org/stable/gallery/color/colormap_reference.html).

            figsize (tuple): 
                The matplotlib figure size to use when displaying, tuple of two integers (ie. `figsize=(14, 4)`)
        
            aspect (str or float): 
                The matplotlib imshow aspect ration to use. A common value for this is `auto`. All valid values 
                can be found on the [matplotlib documentation](https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.imshow.html).

            colorbar (bool): 
                Display a colorbar. Default is `False`.

            title (str): 
                A title to display above the rendered image. Defaults to no title.

            returnfig (bool): 
                Instead of displaying the image, return the matplotlib figure object. This allows for further plot 
                manipulation, for example, adding labels or a title in a different location than the default. 
                
                Remember - if this parameter is supplied, be sure that you close your plot after finishing work 
                with it. This can be achieved by doing `plt.close(fig)`. 
                
                Note that this method cannot be used in combination with `savefig`.

            savefig (bool): 
                Save the displayed image to disk instead of displaying it. The parameter savefig_filename is required if 
                this parameter is set to True. Defaults to `False`.

            savefig_filename (str): 
                Filename to save the image to. Must be specified if the savefig parameter is set to True.

            savefig_quality (int): 
                Quality level of the saved image. This can be specified if the savefig_filename is a JPG image. If it
                is a PNG, quality is ignored. Default quality level for JPGs is matplotlib/Pillow's default of 75%.

        Returns:
            The displayed image, by default. If `savefig` is set to True, nothing will be returned. If `returnfig` is 
            set to True, the plotting variables `(fig, ax)` will be returned.

        Raises:
            ValueError: issues encountered with supplied parameters.
        """
        return func_display(image, cmap, figsize, aspect, colorbar, title, returnfig, savefig, savefig_filename, savefig_quality)

    def movie(
        self,
        input_filenames: List[str],
        output_filename: str,
        n_parallel: int = 1,
        fps: int = 25,
        progress_bar_disable: bool = False,
    ) -> None:
        """
        Generate a movie file from a list of filenames. Note that the codec used is "mp4v".

        Args:
            input_filenames (List[str]): 
                Filenames of frames to use for movie generation. No sorting is applied, so it is 
                assumed the list is in the desired order. This parameter is required.
            
            output_filename (str): 
                Filename for the created movie file. This parameter is required.

            n_parallel (int): 
                Number of multiprocessing workers to use. Default is `1`, which does not use
                multiprocessing.

            fps (int): 
                Frames per second (FPS) for the movie file. Default is `25`.

            progress_bar_disable (bool): 
                Toggle the progress bars off. Default is `False`.        

        Raises:
            IOError: I/O related issue while generating movie
        """
        return func_movie(self.__aurorax_obj, input_filenames, output_filename, n_parallel, fps, progress_bar_disable)

    def scale_intensity(
        self,
        data: np.ndarray,
        min: Optional[float] = None,
        max: Optional[float] = None,
        top: Optional[float] = None,
        memory_saver: bool = True,
    ) -> np.ndarray:
        """
        Scale all values of an array that lie in the range min<=x<=max in to 
        the range 0<=x<=top.
        
        Args:
            data (numpy.ndarray): 
                Data array, can be 2, 3, or 4-dimensional. Assumed to be an image, or array of 
                images. Also assumed that the first 2 dimensions are the image's x and y 
                coordinates, and the following dimensions are some combination of the number of 
                images, and/or the colour channel.

            min (float): 
                Minimum value of array to be considered

            max (float): 
                Maximum value of array to be considered

            top (float): 
                Maximum value of the scaled result. If not supplied, the max value
                of the data array's dtype is used.

            memory_saver (bool): 
                Utilize less RAM when scaling a set of images. Defaults to `True`. If set to `False` then
                the scaling routine will be faster, but will utilize significantly more RAM.

        Returns:
            A new `numpy.ndarray` that is the same dimensions as the inputted data array, 
            with the scaling applied.

        Raises:
            ValueError: issues encountered with supplied min, max, or top value(s)
        """
        return func_scale_intensity(data, min, max, top, memory_saver)

    def set_theme(self, theme: str) -> None:
        """
        A handy wrapper for setting the matplotlib global theme. Common choices are `light`, 
        `dark`, or `default`.

        Args:
            theme (str): 
                Theme name. Common choices are `light`, `dark`, or `default`. If default, then
                matplotlib theme settings will be fully reset to their defaults.

                Additional themes can be found on the 
                [matplotlib documentation](https://matplotlib.org/stable/gallery/style_sheets/style_sheets_reference.html)
        """
        return func_set_theme(theme)



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Class definition for a data product search
"""

from __future__ import annotations
import datetime
from typing import TYPE_CHECKING, Dict, List, Optional, Literal, Union
from .data_product import DataProductData
from ...metadata_filters import MetadataFilter
from ...api import AuroraXAPIRequest
from ...sources import DataSource, FORMAT_BASIC_INFO
from ....exceptions import AuroraXAPIError
from ...requests._requests import (
    cancel as requests_cancel,
    wait_for_data as requests_wait_for_data,
    get_data as requests_get_data,
    get_status as requests_get_status,
)
from ...._util import show_warning
if TYPE_CHECKING:
    from ....pyaurorax import PyAuroraX  # pragma: nocover-ok


class DataProductSearch:
    """
    Class representing a data product search

    Attributes:
        start (datetime.datetime): 
            Start timestamp of the search (inclusive)

        end (datetime.datetime): 
            End timestamp of the search (inclusive)

        programs (List[str]): 
            List of program names to search

        platforms (List[str]): 
            List of platform names to search

        instrument_types (List[str]): 
            List of instrument types to search

        data_product_types (List[str]): 
            List of strings describing data product types to filter on e.g. "keogram", defaults 
            to None. Valid options are: `keogram`, `montage`, `movie`, `summary_plot`, and 
            `data_availability`.
    
        metadata_filters (MetadataFilter or List[Dict]): 
            The metadata filters to use when searching, defaults to None

        metadata_filters_logical_operator (str): 
            The logical operator to use when evaluating metadata filters (either `and` or `or`), 
            defaults to `and`. This parameter is deprecated in exchange for passing a 
            MetadataFilter object into the metadata_filters parameter. 

        response_format (Dict): 
            JSON representation of desired data response format

        request (AuroraXResponse): 
            AuroraXResponse object returned when the search is executed

        request_id (str): 
            Unique ID assigned to the request by the AuroraX API

        request_url (str): 
            Unique URL assigned to the request by the AuroraX API

        executed (bool): 
            Indicates if the search has been executed/started

        completed (bool): 
            Indicates if the search has finished

        data_url (str): 
            The URL where data is accessed

        query (Dict): 
            The query for this request as JSON

        status (Dict): 
            The status of the query

        data (List[DataProductData]): 
            The data product records found

        logs (List[Dict]): 
            All log messages outputted by the AuroraX API for this request
    """

    __STANDARD_POLLING_SLEEP_TIME: float = 1.0

    def __init__(self,
                 aurorax_obj: PyAuroraX,
                 start: datetime.datetime,
                 end: datetime.datetime,
                 programs: Optional[List[str]] = None,
                 platforms: Optional[List[str]] = None,
                 instrument_types: Optional[List[str]] = None,
                 data_product_types: Optional[List[Literal["keogram", "montage", "movie", "summary_plot", "data_availability"]]] = None,
                 metadata_filters: Optional[Union[MetadataFilter, List[Dict]]] = None,
                 metadata_filters_logical_operator: Optional[Literal["and", "or", "AND", "OR"]] = None,
                 response_format: Optional[Dict] = None) -> None:

        # show warnings
        if (isinstance(metadata_filters, MetadataFilter) and metadata_filters_logical_operator is not None):
            # logical operator supplied, but MetadataFilter supplied too
            show_warning("Supplying a MetadataFilter object in addition to the metadata_filters_logical_operator " +
                         "parameter is redundant. Only the MetadataFilter object is needed. The " +
                         "metadata_filters_logical_operator parameter will be ignored")

        # set variables using passed in args
        self.__aurorax_obj = aurorax_obj
        self.start = start
        self.end = end
        self.programs = programs
        self.platforms = platforms
        self.instrument_types = instrument_types
        self.data_product_types = data_product_types
        self.metadata_filters = metadata_filters
        self.metadata_filters_logical_operator = "AND" if metadata_filters_logical_operator is None else metadata_filters_logical_operator
        self.response_format = response_format

        # initialize additional variables
        self.request = None
        self.request_id = ""
        self.request_url = ""
        self.executed = False
        self.completed = False
        self.data_url = ""
        self.__query = {}
        self.status = {}
        self.data = []
        self.logs = []

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return "DataProductSearch(executed=%s, completed=%s, request_id='%s')" % (
            self.executed,
            self.completed,
            self.request_id,
        )

    def pretty_print(self):
        """
        A special print output for this class.
        """
        # set status and query strings
        max_len = 80
        status_str = str(self.status)
        query_str = str(self.query)
        if (len(status_str) > max_len):
            status_str = "%s..." % (status_str[0:max_len])
        if (len(query_str) > max_len):
            query_str = "%s..." % (query_str[0:max_len])

        # set results string
        if (self.executed is True):
            if (len(self.data) == 1):
                data_str = "[1 data product result]"
            else:
                data_str = "[%d data product results]" % (len(self.data))
        else:
            data_str = ""

        # set logs string
        if (self.executed is True):
            if (len(self.logs) == 1):  # pragma: nocover-ok
                logs_str = "[1 log message]"
            else:
                logs_str = "[%d log messages]" % (len(self.logs))
        else:
            logs_str = ""

        # print
        print("DataProductSearch:")
        print("  %-13s: %s" % ("executed", self.executed))
        print("  %-13s: %s" % ("completed", self.completed))
        print("  %-13s: %s" % ("request_id", self.request_id))
        print("  %-13s: %s" % ("request", self.request))
        print("  %-13s: %s" % ("request_url", self.request_url))
        print("  %-13s: %s" % ("data_url", self.data_url))
        print("  %-13s: %s" % ("query", query_str))
        print("  %-13s: %s" % ("status", status_str))
        print("  %-13s: %s" % ("data", data_str))
        print("  %-13s: %s" % ("logs", logs_str))

    @property
    def query(self):
        """
        Property for the query value
        """
        # set metadata filter value
        if (self.metadata_filters is None):
            metadata_filters_dict = {}
        elif (isinstance(self.metadata_filters, MetadataFilter) is True):
            # metadata filter is a class
            metadata_filters_dict = self.metadata_filters.to_query_dict()  # type: ignore
        else:
            # metadata filter is a dictionary
            metadata_filters_dict = {
                "expressions": self.metadata_filters,
                "logical_operator": self.metadata_filters_logical_operator,
            }

        # set query
        self.__query = {
            "data_sources": {
                "programs": [] if not self.programs else self.programs,
                "platforms": [] if not self.platforms else self.platforms,
                "instrument_types": [] if not self.instrument_types else self.instrument_types,
                "data_product_metadata_filters": metadata_filters_dict,
            },
            "start": self.start.strftime("%Y-%m-%dT%H:%M:%S"),
            "end": self.end.strftime("%Y-%m-%dT%H:%M:%S"),
            "data_product_type_filters": [] if not self.data_product_types else self.data_product_types,
        }

        # return
        return self.__query

    def execute(self) -> None:
        """
        Initiate a data product search request
        """
        # do request
        url = "%s/%s" % (self.__aurorax_obj.api_base_url, self.__aurorax_obj.search.api.URL_SUFFIX_DATA_PRODUCTS_SEARCH)
        req = AuroraXAPIRequest(self.__aurorax_obj, method="post", url=url, body=self.query, null_response=True)
        res = req.execute()

        # set request ID, request_url, executed
        self.executed = True
        if (res.status_code == 202):
            # request successfully dispatched
            self.executed = True
            self.request_url = res.request.headers["location"]
            self.request_id = self.request_url.rsplit("/", 1)[-1]

        # set request variable
        self.request = res

    def update_status(self, status: Optional[Dict] = None) -> None:
        """
        Update the status of this data product search request

        Args:
            status (Dict): 
                The previously-retrieved status of this request (include
                to avoid requesting it from the API again), defaults to None
        """
        # get the status if it isn't passed in
        if (status is None):
            status = requests_get_status(self.__aurorax_obj, self.request_url)

        # check response
        if (status is None):  # pragma: nocover-ok
            raise AuroraXAPIError("Could not retrieve status for this request")

        # update request status by checking if data URI is set
        if (status["search_result"]["data_uri"] is not None):
            self.completed = True
            self.data_url = "%s/data" % (self.request_url)

        # set class variable "status" and "logs"
        self.status = status
        self.logs = status["logs"]

    def check_for_data(self) -> bool:
        """
        Check to see if data is available for this data product
        search request

        Returns:
            True if data is available, else False
        """
        self.update_status()
        return self.completed

    def get_data(self) -> None:
        """
        Retrieve the data available for this data product search request
        """
        # check if completed yet
        if (self.completed is False):
            print("No data available, update status or check for data first")
            return

        # get data
        raw_data = requests_get_data(self.__aurorax_obj, self.data_url, self.response_format, False)

        # set data variable
        if (self.response_format is not None):
            self.data = raw_data
        else:
            # cast data source objects
            for i in range(0, len(raw_data)):
                ds = DataSource(**raw_data[i]["data_source"], format=FORMAT_BASIC_INFO)
                raw_data[i]["data_source"] = ds

            # cast data product objects
            self.data = [DataProductData(**dp) for dp in raw_data]

    def wait(self, poll_interval: float = __STANDARD_POLLING_SLEEP_TIME, verbose: bool = False) -> None:
        """
        Block and wait for the request to complete and data is available
        for retrieval

        Args:
            poll_interval (float): 
                Time in seconds to wait between polling attempts, defaults to 1 second

            verbose (bool): 
                Output poll times and other progress messages, defaults to False
        """
        url = "%s/%s" % (self.__aurorax_obj.api_base_url, self.__aurorax_obj.search.api.URL_SUFFIX_DATA_PRODUCTS_REQUEST.format(self.request_id))
        self.update_status(requests_wait_for_data(self.__aurorax_obj, url, poll_interval, verbose))

    def cancel(self, wait: bool = False, poll_interval: float = __STANDARD_POLLING_SLEEP_TIME, verbose: bool = False) -> int:
        """
        Cancel the data product search request

        This method returns immediately by default since the API processes
        this request asynchronously. If you would prefer to wait for it
        to be completed, set the 'wait' parameter to True. You can adjust
        the polling time using the 'poll_interval' parameter.

        Args:
            wait (bool): 
                Wait until the cancellation request has been completed (may wait 
                for several minutes)
            
            poll_interval (float): 
                Seconds to wait between polling calls, defaults to 1 second.
            
            verbose (bool): 
                Output poll times and other progress messages, defaults to False

        Returns:
            1 on success

        Raises:
            pyaurorax.exceptions.AuroraXAPIError: An API error was encountered
            pyaurorax.exceptions.AuroraXUnauthorizedError: Invalid API key for this operation
        """
        url = "%s/%s" % (self.__aurorax_obj.api_base_url, self.__aurorax_obj.search.api.URL_SUFFIX_DATA_PRODUCTS_REQUEST.format(self.request_id))
        return requests_cancel(self.__aurorax_obj, url, wait, poll_interval, verbose)

    def describe(self):
        """
        Describe the data products search as an "SQL-like" string.

        Returns:
            The "SQL-like" string describing the data products search object
        """
        # make request
        url = "%s/%s" % (self.__aurorax_obj.api_base_url, self.__aurorax_obj.search.api.URL_SUFFIX_DESCRIBE_DATA_PRODUCTS_QUERY)
        req = AuroraXAPIRequest(self.__aurorax_obj, method="post", url=url, body=self.query)
        res = req.execute()

        # return
        return res.data



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Class definition for an ephemeris search
"""

from __future__ import annotations
import datetime
from typing import TYPE_CHECKING, Dict, List, Optional, Literal, Union
from ...metadata_filters import MetadataFilter
from .ephemeris import EphemerisData
from ...api import AuroraXAPIRequest
from ...sources.classes.data_source import DataSource, FORMAT_BASIC_INFO
from ....exceptions import AuroraXError, AuroraXAPIError
from ...requests._requests import (
    cancel as requests_cancel,
    wait_for_data as requests_wait_for_data,
    get_data as requests_get_data,
    get_status as requests_get_status,
)
from ...._util import show_warning
if TYPE_CHECKING:
    from ....pyaurorax import PyAuroraX  # pragma: nocover-ok


class EphemerisSearch:
    """
    Class representing an ephemeris search

    Note: At least one search criteria from programs, platforms, or instrument_types
    must be specified.

    Attributes:
        start (datetime.datetime): 
            Start timestamp of the search (inclusive)

        end (datetime.datetime): 
            End timestamp of the search (inclusive)

        programs (List[str]): 
            List of programs to search through, defaults to None

        platforms (List[str]): 
            List of platforms to search through, defaults to None

        instrument_types (List[str]): 
            List of instrument types to search through, defaults to None

        metadata_filters (MetadataFilter or List[Dict]): 
            List of dictionaries describing metadata keys and values to filter on, defaults 
            to None

        metadata_filters_logical_operator (str): 
            The logical operator to use when evaluating metadata filters (either `AND` or `OR`), 
            defaults to `AND`

        response_format (Dict): 
            JSON representation of desired data response format
        
        request (AuroraXResponse): 
            AuroraXResponse object returned when the search is executed
        
        request_id (str): 
            Unique ID assigned to the request by the AuroraX API
       
        request_url (str): 
            Unique URL assigned to the request by the AuroraX API
     
        executed (bool): 
            Indicates if the search has been executed/started
     
        completed (bool): 
            Indicates if the search has finished
       
        data_url (str): 
            The URL where data is accessed
      
        query (Dict): 
            The query for this request as JSON
      
        status (Dict): 
            The status of the query
      
        data (List[EphemerisData]): 
            The ephemeris records found
      
        logs (List[Dict]): 
            All log messages outputted by the AuroraX API for this request
    """

    __STANDARD_POLLING_SLEEP_TIME: float = 1.0

    def __init__(self,
                 aurorax_obj: PyAuroraX,
                 start: datetime.datetime,
                 end: datetime.datetime,
                 programs: Optional[List[str]] = None,
                 platforms: Optional[List[str]] = None,
                 instrument_types: Optional[List[str]] = None,
                 metadata_filters: Optional[Union[MetadataFilter, List[Dict]]] = None,
                 metadata_filters_logical_operator: Optional[Literal["and", "or", "AND", "OR"]] = None,
                 response_format: Optional[Dict] = None) -> None:

        # show warnings
        if (isinstance(metadata_filters, MetadataFilter) and metadata_filters_logical_operator is not None):
            # logical operator supplied, but MetadataFilter supplied too
            show_warning("Supplying a MetadataFilter object in addition to the metadata_filters_logical_operator " +
                         "parameter is redundant. Only the MetadataFilter object is needed. The " +
                         "metadata_filters_logical_operator parameter will be ignored")

        # set variables using passed in args
        self.__aurorax_obj = aurorax_obj
        self.start = start
        self.end = end
        self.programs = programs
        self.platforms = platforms
        self.instrument_types = instrument_types
        self.metadata_filters = metadata_filters
        self.metadata_filters_logical_operator = "AND" if metadata_filters_logical_operator is None else metadata_filters_logical_operator.upper()
        self.response_format = response_format

        # initialize additional variables
        self.request = None
        self.request_id = ""
        self.request_url = ""
        self.executed = False
        self.completed = False
        self.data_url = ""
        self.__query = {}
        self.status = {}
        self.data = []
        self.logs = []

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return "EphemerisSearch(executed=%s, completed=%s, request_id='%s')" % (
            self.executed,
            self.completed,
            self.request_id,
        )

    def pretty_print(self):
        """
        A special print output for this class.
        """
        # set status and query strings
        max_len = 80
        status_str = str(self.status)
        query_str = str(self.query)
        if (len(status_str) > max_len):
            status_str = "%s..." % (status_str[0:max_len])
        if (len(query_str) > max_len):
            query_str = "%s..." % (query_str[0:max_len])

        # set results string
        if (self.executed is True):
            if (len(self.data) == 1):
                data_str = "[1 ephemeris result]"
            else:
                data_str = "[%d ephemeris results]" % (len(self.data))
        else:
            data_str = ""

        # set logs string
        if (self.executed is True):
            if (len(self.logs) == 1):  # pragma: nocover-ok
                logs_str = "[1 log message]"
            else:
                logs_str = "[%d log messages]" % (len(self.logs))
        else:
            logs_str = ""

        # print
        print("EphemerisSearch:")
        print("  %-13s: %s" % ("executed", self.executed))
        print("  %-13s: %s" % ("completed", self.completed))
        print("  %-13s: %s" % ("request_id", self.request_id))
        print("  %-13s: %s" % ("request", self.request))
        print("  %-13s: %s" % ("request_url", self.request_url))
        print("  %-13s: %s" % ("data_url", self.data_url))
        print("  %-13s: %s" % ("query", query_str))
        print("  %-13s: %s" % ("status", status_str))
        print("  %-13s: %s" % ("data", data_str))
        print("  %-13s: %s" % ("logs", logs_str))

    @property
    def query(self):
        """
        Property for the query value
        """
        # set metadata filter value
        if (self.metadata_filters is None):
            metadata_filters_dict = {}
        elif (isinstance(self.metadata_filters, MetadataFilter) is True):
            # metadata filter is a class
            metadata_filters_dict = self.metadata_filters.to_query_dict()  # type: ignore
        else:
            # metadata filter is a dictionary
            metadata_filters_dict = {
                "expressions": self.metadata_filters,
                "logical_operator": self.metadata_filters_logical_operator,
            }

        # set query
        self.__query = {
            "data_sources": {
                "programs": [] if not self.programs else self.programs,
                "platforms": [] if not self.platforms else self.platforms,
                "instrument_types": [] if not self.instrument_types else self.instrument_types,
                "ephemeris_metadata_filters": metadata_filters_dict,
            },
            "start": self.start.strftime("%Y-%m-%dT%H:%M:%S"),
            "end": self.end.strftime("%Y-%m-%dT%H:%M:%S"),
        }

        # return
        return self.__query

    def execute(self) -> None:
        """
        Initiate ephemeris search request

        Raises:
            pyaurorax.exceptions.AuroraXError: Invalid request parameters are set
        """
        # check for at least one filter criteria
        if not (self.programs or self.platforms or self.instrument_types or self.metadata_filters):  # pragma: nocover-ok
            raise AuroraXError("At least one filter criteria parameter besides 'start' and 'end' must be specified")

        # do request
        url = "%s/%s" % (self.__aurorax_obj.api_base_url, self.__aurorax_obj.search.api.URL_SUFFIX_EPHEMERIS_SEARCH)
        req = AuroraXAPIRequest(self.__aurorax_obj, method="post", url=url, body=self.query, null_response=True)
        res = req.execute()

        # set request ID, request_url, executed
        self.executed = True
        if (res.status_code == 202):
            # request successfully dispatched
            self.executed = True
            self.request_url = res.request.headers["location"]
            self.request_id = self.request_url.rsplit("/", 1)[-1]

        # set the request variable
        self.request = res

    def update_status(self, status: Optional[Dict] = None) -> None:
        """
        Update the status of this ephemeris search request

        Args:
            status (Dict): 
                The previously-retrieved status of this request (include to avoid requesting it 
                from the API again), defaults to None
        """
        # get the status if it isn't passed in
        if (status is None):
            status = requests_get_status(self.__aurorax_obj, self.request_url)

        # check response
        if (status is None):  # pragma: nocover-ok
            raise AuroraXAPIError("Could not retrieve status for this request")

        # update request status by checking if data URI is set
        if (status["search_result"]["data_uri"] is not None):
            self.completed = True
            self.data_url = "%s/data" % (self.request_url)

        # set class variable "status" and "logs"
        self.status = status
        self.logs = status["logs"]

    def check_for_data(self) -> bool:
        """
        Check to see if data is available for this ephemeris search request

        Returns:
            True if data is available, else False
        """
        self.update_status()
        return self.completed

    def get_data(self) -> None:
        """
        Retrieve the data available for this ephemeris search request
        """
        # check if completed yet
        if (self.completed is False):
            print("No data available, update status or check for data first")
            return

        # get data
        raw_data = requests_get_data(self.__aurorax_obj, self.data_url, self.response_format, False)

        # set data variable
        if (self.response_format is not None):
            self.data = raw_data
        else:
            # cast data source objects
            for i in range(0, len(raw_data)):
                ds = DataSource(**raw_data[i]["data_source"], format=FORMAT_BASIC_INFO)
                raw_data[i]["data_source"] = ds

            # cast ephemeris objects
            self.data = [EphemerisData(**e) for e in raw_data]

    def wait(self, poll_interval: float = __STANDARD_POLLING_SLEEP_TIME, verbose: bool = False) -> None:
        """
        Block and wait for the request to complete and data is available for retrieval

        Args:
            poll_interval (float): 
                Time in seconds to wait between polling attempts, defaults to 1 second

            verbose (bool): 
                Output poll times and other progress messages, defaults to False
        """
        url = "%s/%s" % (self.__aurorax_obj.api_base_url, self.__aurorax_obj.search.api.URL_SUFFIX_EPHEMERIS_REQUEST.format(self.request_id))
        self.update_status(requests_wait_for_data(self.__aurorax_obj, url, poll_interval, verbose))

    def cancel(self, wait: bool = False, poll_interval: float = __STANDARD_POLLING_SLEEP_TIME, verbose: bool = False) -> int:
        """
        Cancel the ephemeris search request

        This method returns immediately by default since the API processes
        this request asynchronously. If you would prefer to wait for it
        to be completed, set the 'wait' parameter to True. You can adjust
        the polling time using the 'poll_interval' parameter.

        Args:
            wait (bool): 
                Wait until the cancellation request has been completed (may wait for 
                several minutes)

            poll_interval (float): 
                Seconds to wait between polling calls, defaults to 1 second

            verbose (bool): 
                Output poll times and other progress messages, defaults to False

        Returns:
            1 on success

        Raises:
            pyaurorax.exceptions.AuroraXUnauthorizedError: Invalid API key for this operation
            pyaurorax.exceptions.AuroraXAPIError: An API error was encountered
        """
        url = "%s/%s" % (self.__aurorax_obj.api_base_url, self.__aurorax_obj.search.api.URL_SUFFIX_EPHEMERIS_REQUEST.format(self.request_id))
        return requests_cancel(self.__aurorax_obj, url, wait, poll_interval, verbose)

    def describe(self):
        """
        Describe the ephemeris search as an "SQL-like" string.

        Returns:
            The "SQL-like" string describing the ephemeris search object
        """
        # make request
        url = "%s/%s" % (self.__aurorax_obj.api_base_url, self.__aurorax_obj.search.api.URL_SUFFIX_DESCRIBE_EPHEMERIS_QUERY)
        req = AuroraXAPIRequest(self.__aurorax_obj, method="post", url=url, body=self.query)
        res = req.execute()

        # return
        return res.data



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Calibration procedures for REGO data.
"""

from ._common import (
    perform_dark_frame_correction,
    perform_flatfield_calibration,
    perform_rayleighs_correction,
)


def apply_calibration(
    images,
    cal_flatfield,
    cal_rayleighs,
    step_dark_frame_correction,
    step_flatfield_calibration,
    step_rayleighs_calibration,
    exposure_length_sec,
):
    # verify that we have everything we need for each requested step
    if (step_flatfield_calibration is True and cal_flatfield is None):
        raise ValueError("The cal_flatfield parameter must be supplied to perform the flatfield correction step")
    if (step_rayleighs_calibration is True and cal_rayleighs is None):
        raise ValueError("The cal_rayleighs parameter must be supplied to perform the rayleighs conversion step")

    # init
    calibrated_images = images

    # perform the dark frame correction
    #
    # NOTE: we do a 5x5 bottom corner mean
    if (step_dark_frame_correction is True):
        calibrated_images = perform_dark_frame_correction(images, 5)

    # apply the flatfield correction
    if (step_flatfield_calibration is True and cal_flatfield is not None):
        calibrated_images = perform_flatfield_calibration(calibrated_images, cal_flatfield)

    # apply the rayleighs conversion
    if (step_rayleighs_calibration is True and cal_rayleighs is not None):
        calibrated_images = perform_rayleighs_correction(calibrated_images, cal_rayleighs, exposure_length_sec)

    # return
    return calibrated_images



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Manage AuroraX data sources utilized by the search engine.
"""

from typing import Optional, List, Dict
from texttable import Texttable
from ._sources import list as func_list
from ._sources import search as func_search
from ._sources import get as func_get
from ._sources import get_using_filters as func_get_using_filters
from ._sources import get_using_identifier as func_get_using_identifier
from ._sources import add as func_add
from ._sources import delete as func_delete
from ._sources import update as func_update
from .classes.data_source import DataSource, DataSourceStatistics
from .classes.data_source import (
    FORMAT_BASIC_INFO,
    FORMAT_BASIC_INFO_WITH_METADATA,
    FORMAT_DEFAULT,
    FORMAT_FULL_RECORD,
    FORMAT_IDENTIFIER_ONLY,
    SOURCE_TYPE_EVENT_LIST,
    SOURCE_TYPE_GROUND,
    SOURCE_TYPE_HEO,
    SOURCE_TYPE_LEO,
    SOURCE_TYPE_LUNAR,
    SOURCE_TYPE_NOT_APPLICABLE,
)

__all__ = [
    "SourcesManager",
    "DataSource",
    "DataSourceStatistics",
    "FORMAT_BASIC_INFO",
    "FORMAT_BASIC_INFO_WITH_METADATA",
    "FORMAT_DEFAULT",
    "FORMAT_FULL_RECORD",
    "FORMAT_IDENTIFIER_ONLY",
    "SOURCE_TYPE_EVENT_LIST",
    "SOURCE_TYPE_GROUND",
    "SOURCE_TYPE_HEO",
    "SOURCE_TYPE_LEO",
    "SOURCE_TYPE_LUNAR",
    "SOURCE_TYPE_NOT_APPLICABLE",
]


class SourcesManager:
    """
    The SourcesManager object is initialized within every PyAuroraX object. It acts as a way to access 
    the submodules and carry over configuration information in the super class.
    """

    def __init__(self, aurorax_obj):
        self.__aurorax_obj = aurorax_obj

    def list(self,
             program: Optional[str] = None,
             platform: Optional[str] = None,
             instrument_type: Optional[str] = None,
             source_type: Optional[str] = None,
             owner: Optional[str] = None,
             format: Optional[str] = FORMAT_FULL_RECORD,
             order: Optional[str] = "identifier",
             include_stats: Optional[bool] = False) -> List[DataSource]:
        """
        Retrieve all data source records. Parameters can be used to filter as desired.

        Args:
            program (str): 
                The program to filter for, defaults to `None`

            platform (str): 
                The platform to filter for, defaults to `None`

            instrument_type (str): 
                The instrument type to filter for, defaults to `None`

            source_type (str): 
                The data source type to filter for, defaults to `None`. Options are in 
                the pyaurorax.search.sources module, or at the top level using the 
                pyaurorax.search.SOURCE_TYPE_* variables.
                
            owner (str): 
                The owner's email address to filter for, defaults to `None`

            format (str): 
                The format of the data sources returned, defaults to `classes.data_source.FORMAT_FULL_RECORD`. 
                Other options are in the pyaurorax.search.sources module, or at the top level using 
                the pyaurorax.search.FORMAT_* variables.

            order (str): 
                The category to order results by. Valid values are identifier, program, platform,
                instrument_type, display_name, or owner. Defaults to `identifier`

            include_stats (bool): 
                Include additional stats information about the data source, defaults to `False`

        Returns:
            A list of `DataSource` records matching the requested parameters

        Raises:
            pyaurorax.exceptions.AuroraXAPIError: Error during API call
        """
        return func_list(
            self.__aurorax_obj,
            program,
            platform,
            instrument_type,
            source_type,
            owner,
            format,
            order,
            include_stats,
        )

    def list_in_table(self,
                      program: Optional[str] = None,
                      platform: Optional[str] = None,
                      instrument_type: Optional[str] = None,
                      source_type: Optional[str] = None,
                      owner: Optional[str] = None,
                      order: Optional[str] = "identifier",
                      table_max_width: int = 200,
                      limit: Optional[int] = None) -> None:
        """
        Display all data source records in a table. Parameters can be used to filter as desired.

        Args:
            program (str): 
                The program to filter for, defaults to `None`

            platform (str): 
                The platform to filter for, defaults to `None`

            instrument_type (str): 
                The instrument type to filter for, defaults to `None`

            source_type (str): 
                The data source type to filter for, defaults to `None`. Options are in 
                the pyaurorax.search.sources module, or at the top level using the 
                pyaurorax.search.SOURCE_TYPE_* variables.

            owner (str): 
                The owner's email address to filter for, defaults to `None`

            format (str): 
                The format of the data sources returned, defaults to `classes.data_source.FORMAT_FULL_RECORD`. 
                Other options are in the pyaurorax.search.sources module, or at the top level using 
                the pyaurorax.search.FORMAT_* variables.

            order (str): 
                The category to order results by. Valid values are identifier, program, platform,
                instrument_type, display_name, or owner. Defaults to `identifier`
                
            table_max_width (int): 
                Table maximum width, defaults to 200

            limit (int): 
                Limit the table rows to a certain value, regardless of how many sources it found
                to display

        Returns:
            No return, only prints a table

        Raises:
            pyaurorax.exceptions.AuroraXAPIError: Error during API call
        """
        # get datasets
        datasets = func_list(
            self.__aurorax_obj,
            program,
            platform,
            instrument_type,
            source_type,
            owner,
            FORMAT_FULL_RECORD,
            order,
            False,
        )

        # reduce based on limit
        if (limit is not None and len(datasets) > limit):
            datasets = datasets[0:limit]

        # set table lists
        table_identifiers = []
        table_programs = []
        table_platforms = []
        table_instrument_types = []
        table_source_types = []
        table_display_names = []
        for d in datasets:
            table_identifiers.append(d.identifier)
            table_programs.append(d.program)
            table_platforms.append(d.platform)
            table_instrument_types.append(d.instrument_type)
            table_source_types.append(d.source_type)
            table_display_names.append(d.display_name)

        # set header values
        table_headers = [
            "Identifier",
            "Program",
            "Platform",
            "Instrument Type",
            "Source Type",
            "Display Name",
        ]

        # print as table
        table = Texttable(max_width=table_max_width)
        table.set_deco(Texttable.HEADER)
        table.set_cols_dtype(["t"] * len(table_headers))
        table.set_header_align(["l"] * len(table_headers))
        table.set_cols_align(["l"] * len(table_headers))
        table.header(table_headers)
        for i in range(0, len(table_identifiers)):
            table.add_row([
                table_identifiers[i],
                table_programs[i],
                table_platforms[i],
                table_instrument_types[i],
                table_source_types[i],
                table_display_names[i],
            ])
        print(table.draw())

    def search(self,
               programs: Optional[List[str]] = [],
               platforms: Optional[List[str]] = [],
               instrument_types: Optional[List[str]] = [],
               format: Optional[str] = FORMAT_FULL_RECORD,
               order: Optional[str] = "identifier",
               include_stats: Optional[bool] = False) -> List[DataSource]:
        """
        Search for data source records. Parameters can be used to filter as desired.

        This function is very similar to the `list()` function, however multiple programs,
        platforms, and/or instrument types can be supplied here. The `list()` function only 
        supports single values for the parameters.

        Args:
            programs (List[str]): 
                The programs to search for, defaults to `[]`

            platforms (List[str]): 
                The platforms to search for, defaults to `[]`

            instrument_types (List[str]): 
                The instrument types to search for, defaults to `[]`

            format (str): 
                The format of the data sources returned, defaults to `classes.data_source.FORMAT_FULL_RECORD`. 
                Other options are in the pyaurorax.search.sources module, or at the top level using 
                the pyaurorax.search.FORMAT_* variables.

            order (str): 
                The category to order results by. Valid values are identifier, program, platform,
                instrument_type, display_name, or owner. Defaults to `identifier`

            include_stats (bool): 
                Include additional stats information about the data source, defaults to `False`

        Returns:
            A list of `DataSource` records matching the requested parameters

        Raises:
            pyaurorax.exceptions.AuroraXAPIError: Error during API call
        """
        return func_search(
            self.__aurorax_obj,
            programs,
            platforms,
            instrument_types,
            format,
            order,
            include_stats,
        )

    def get(self,
            program: str,
            platform: str,
            instrument_type: str,
            format: Optional[str] = FORMAT_FULL_RECORD,
            include_stats: Optional[bool] = False) -> DataSource:
        """
        Retrieve a specific data source record

        Args:
            program (str): 
                The program name

            platform (str): 
                The platform name

            instrument_type (str): 
                The instrument type name

            format (str): 
                The format of the data sources returned, defaults to `classes.data_source.FORMAT_FULL_RECORD`. 
                Other options are in the pyaurorax.search.sources module, or at the top level using 
                the pyaurorax.search.FORMAT_* variables.

            include_stats (bool): 
                Include additional stats information about the data source, defaults to `False`

        Returns:
            The `DataSource` matching the requested parameters

        Raises:
            pyaurorax.exceptions.AuroraXAPIError: Error during API call
            pyaurorax.exceptions.AuroraXNotFoundError: Source not found
        """
        return func_get(
            self.__aurorax_obj,
            program,
            platform,
            instrument_type,
            format,
            include_stats,
        )

    def get_using_filters(self,
                          program: Optional[str] = None,
                          platform: Optional[str] = None,
                          instrument_type: Optional[str] = None,
                          source_type: Optional[str] = None,
                          owner: Optional[str] = None,
                          format: Optional[str] = FORMAT_FULL_RECORD,
                          order: Optional[str] = "identifier",
                          include_stats: Optional[bool] = False) -> List[DataSource]:
        """
        Retrieve all data sources matching a filter

        Args:
            program (str): 
                The program to filter for, defaults to `None`

            platform (str): 
                The platform to filter for, defaults to `None`
                
            instrument_type (str): 
                The instrument type to filter for, defaults to `None`

            source_type (str): 
                The data source type to filter for, defaults to `None`. Options are in 
                the pyaurorax.search.sources module, or at the top level using the 
                pyaurorax.search.SOURCE_TYPE_* variables.

            owner (str): 
                The owner's email address to filter for, defaults to `None`

            format (str): 
                The format of the data sources returned, defaults to `classes.data_source.FORMAT_FULL_RECORD`. 
                Other options are in the pyaurorax.search.sources module, or at the top level using 
                the pyaurorax.search.FORMAT_* variables.

            order (str): 
                The category to order results by. Valid values are identifier, program, platform,
                instrument_type, display_name, or owner. Defaults to `identifier`

            include_stats (bool): 
                Include additional stats information about the data source, defaults to `False`.

        Returns:
            A list of `DataSource` records matching the requested parameters

        Raises:
            pyaurorax.exceptions.AuroraXAPIError: Error during API call
        """
        return func_get_using_filters(
            self.__aurorax_obj,
            program,
            platform,
            instrument_type,
            source_type,
            owner,
            format,
            order,
            include_stats,
        )

    def get_using_identifier(self, identifier: int, format: Optional[str] = FORMAT_FULL_RECORD, include_stats: Optional[bool] = False) -> DataSource:
        """
        Retrieve data source for a specific identifier

        Args:
            identifier (int): 
                The AuroraX unique data source identifier number

            format (str): 
                The format of the data sources returned, defaults to `classes.data_source.FORMAT_FULL_RECORD`. 
                Other options are in the pyaurorax.search.sources module, or at the top level using 
                the pyaurorax.search.FORMAT_* variables.

            include_stats (bool): 
                Include additional stats information about the data source, defaults to `False`

        Returns:
            The `DataSource` for the specified identifier

        Raises:
            pyaurorax.exceptions.AuroraXAPIError: Error during API call
        """
        return func_get_using_identifier(self.__aurorax_obj, identifier, format, include_stats)

    def add(self, data_source: DataSource) -> DataSource:
        """
        Add a new data source to the AuroraX search engine

        Args:
            data_source (DataSource): 
                The data source to add. THe data source record must have at least the following 
                values specified: program, platform, instrument_type, display_name, and source_type.

        Returns:
            The newly created `DataSource`.

        Raises:
            ValueError: Invalid values for DataSource supplied
            pyaurorax.exceptions.AuroraXAPIError: Error during API call
            pyaurorax.exceptions.AuroraXUnauthorizedError: Not allowed to perform task, or API key / user permissions are invalid
            pyaurorax.exceptions.AuroraXDuplicateError: Duplicate data source, already exists
        """
        return func_add(self.__aurorax_obj, data_source)

    def delete(self, identifier: int) -> int:
        """
        Delete a data source from the AuroraX search engine

        Args:
            identifier (int): 
                The data source unique identifier to delete

        Returns:
            0 on success, raises error if an issue was encountered

        Raises:
            pyaurorax.exceptions.AuroraXAPIError: Error during API call
            pyaurorax.exceptions.AuroraXUnauthorizedError: Not allowed to perform task, or API key / user permissions are invalid
            pyaurorax.exceptions.AuroraXNotFoundError: Data source not found
            pyaurorax.exceptions.AuroraXConflictError: A conflict occurred
        """
        return func_delete(self.__aurorax_obj, identifier)

    def update(self,
               identifier: int,
               program: Optional[str] = None,
               platform: Optional[str] = None,
               instrument_type: Optional[str] = None,
               source_type: Optional[str] = None,
               display_name: Optional[str] = None,
               metadata: Optional[Dict] = None,
               owner: Optional[str] = None,
               maintainers: Optional[List[str]] = None,
               ephemeris_metadata_schema: Optional[List[Dict]] = None,
               data_product_metadata_schema: Optional[List[Dict]] = None) -> DataSource:
        """
        Update a data source in the AuroraX search engine. Omitted fields are ignored during the 
        update. 
        
        Note that the identifier cannot be updated. If you need to update the data source's identifier,
        we recommend deletion of the original data source and recreation using the desired identifier.

        Args:
            identifier (int): 
                The AuroraX unique identifier for the data source, required and cannot be updated

            program (str): 
                The new program for the data source, defaults to `None`

            platform (str): 
                The new platform for the data source, defaults to `None`

            instrument_type (str): 
                The new instrument type for the data source, defaults to `None`

            source_type (str): 
                The new source type for the data source, defaults to `None`. Options
                are in the pyaurorax.search.sources module, or at the top level using the
                pyaurorax.search.SOURCE_TYPE_* variables.

            display_name (str): 
                The new display name for the data source, defaults to `None`

            metadata (Dict): 
                The new metadata for the data source, defaults to `None`

            owner (str): 
                The new owner for the data source, defaults to `None`

            maintainers (str): 
                The new maintainer AuroraX account email addresses, defaults to `None`

            ephemeris_metadata_schema (List[Dict]): 
                A list of dictionaries capturing the metadata keys and values that can 
                appear in ephemeris records associated with the data source, defaults to `None`

            data_product_metadata_schema (List[Dict]): 
                A list of dictionaries capturing the metadata keys and values that can appear 
                in data product records associated with the data source, defaults to `None`

        Returns:
            The updated `DataSource` record

        Raises:
            pyaurorax.exceptions.AuroraXAPIError: Error during API call
            pyaurorax.exceptions.AuroraXUnauthorizedError: Not allowed to perform task, or API key / user permissions are invalid
            pyaurorax.exceptions.AuroraXNotFoundError: Data source not found
        """
        return func_update(
            self.__aurorax_obj,
            identifier,
            program,
            platform,
            instrument_type,
            source_type,
            display_name,
            metadata,
            owner,
            maintainers,
            ephemeris_metadata_schema,
            data_product_metadata_schema,
        )



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import click
import pprint
import json
import os
import datetime
import pyaurorax
from dateutil.parser import parse
from ..helpers import (
    print_request_logs_table,
    print_request_status,
    get_search_data,
)
from ....search.api import URL_SUFFIX_CONJUNCTION_REQUEST
from ...templates import CONJUNCTION_SEARCH_TEMPLATE

# globals
__STANDARD_POLLING_SLEEP_TIME: float = 1.0
__BROWSER_TYPES = sorted([
    "default",
    "mozilla",
    "firefox",
    "netscape",
    "galeon",
    "epiphany",
    "skipstone",
    "kfmclient",
    "konqueror",
    "kfm",
    "mosaic",
    "opera",
    "grail",
    "links",
    "elinks",
    "lynx",
    "w3m",
    "windows-default",
    "macosx",
    "safari",
    "google-chrome",
    "chrome",
    "chromium",
    "chromium-browser",
])


def __create_search_object_from_query(aurorax_obj, q):
    start = parse(q["start"], ignoretz=True)
    end = parse(q["end"], ignoretz=True)
    max_distances = q["max_distances"]
    ground = [] if "ground" not in q else q["ground"]
    space = [] if "space" not in q else q["space"]
    events = [] if "events" not in q else q["events"]
    custom = [] if "adhoc" not in q else q["adhoc"]
    conjunction_types = [] if "conjunction_types" not in q else q["conjunction_types"]
    s = pyaurorax.search.ConjunctionSearch(aurorax_obj,
                                           start,
                                           end,
                                           distance=max_distances,
                                           ground=ground,
                                           space=space,
                                           events=events,
                                           custom_locations=custom,
                                           conjunction_types=conjunction_types)
    return s


@click.group("conjunctions", help="Interact with conjunction searches")
def conjunctions_group():
    pass


@conjunctions_group.command("get_status", short_help="Get status info for a conjunction search request")
@click.argument("request_uuid", type=str)
@click.option("--show-logs", "show_logs", is_flag=True, help="Show the logs for the request")
@click.option("--show-query", "show_query", is_flag=True, help="Show the query for the request")
@click.option("--filter-logs", type=click.Choice(["debug", "info", "warn", "error"]), help="Filter log messages (used with --show-logs)")
@click.option("--table-max-width", "--max-width", type=int, help="Max width for the logs table")
@click.pass_obj
def get_status(config, request_uuid, show_logs, show_query, filter_logs, table_max_width):
    """
    Get information for a conjunction search request

    \b
    REQUEST_UUID    the request unique identifier
    """
    # get request status
    try:
        url = "%s/%s" % (config.aurorax.api_base_url, URL_SUFFIX_CONJUNCTION_REQUEST.format(request_uuid))
        s = config.aurorax.search.requests.get_status(url)
    except pyaurorax.AuroraXAPIError as e:
        if ("API error code 404" in str(e)):
            click.echo("%s occurred: request ID not found" % (type(e).__name__))
        else:  # pragma: nocover-ok
            click.echo("%s occurred: unexpected error -- %s" % (type(e).__name__, e.args[0]))
        sys.exit(1)
    except pyaurorax.AuroraXError as e:  # pragma: nocover-ok
        click.echo("%s occurred: %s" % (type(e).__name__, e.args[0]))
        sys.exit(1)

    # print status nicely
    print_request_status(
        config.aurorax,
        s,
        show_logs=show_logs,
        show_query=show_query,
        filter_logs=filter_logs,
        table_max_width=table_max_width,
    )


@conjunctions_group.command("get_logs", short_help="Get logs for a conjunction search request")
@click.argument("request_uuid", type=str)
@click.option("--filter", "--filter-logs", "filter_", type=click.Choice(["debug", "info", "warn", "error"]), help="Filter log messages")
@click.option("--table-max-width", "--max-width", type=int, help="Max width for the logs table")
@click.option("--no-truncate", is_flag=True, help="Do not truncate log messages that are >1000 characters long")
@click.pass_obj
def get_logs(config, request_uuid, filter_, table_max_width, no_truncate):
    """
    Get the logs for a conjunction search request

    \b
    REQUEST_UUID    the request unique identifier
    """
    # get request status
    try:
        url = "%s/%s" % (config.aurorax.api_base_url, URL_SUFFIX_CONJUNCTION_REQUEST.format(request_uuid))
        s = config.aurorax.search.requests.get_status(url)
    except pyaurorax.AuroraXAPIError as e:
        if ("API error code 404" in str(e)):
            click.echo("%s occurred: request ID not found" % (type(e).__name__))
        else:  # pragma: nocover-ok
            click.echo("%s occurred: unexpected error -- %s" % (type(e).__name__, e.args[0]))
        sys.exit(1)
    except pyaurorax.AuroraXError as e:  # pragma: nocover-ok
        click.echo("%s occurred: %s" % (type(e).__name__, e.args[0]))
        sys.exit(1)

    # set truncate
    truncate = not no_truncate

    # print out the logs nicely
    if ("logs" in s):
        print_request_logs_table(s["logs"], filter_level=filter_, table_max_width=table_max_width, truncate=truncate)
    else:
        click.echo("Search logs: missing, unable to display")  # pragma: nocover


@conjunctions_group.command("get_query", short_help="Get query for a conjunction search request")
@click.argument("request_uuid", type=str)
@click.pass_obj
def get_query(config, request_uuid):
    """
    Get the query for a conjunction search request

    \b
    REQUEST_UUID    the request unique identifier
    """
    # get request status
    try:
        url = "%s/%s" % (config.aurorax.api_base_url, URL_SUFFIX_CONJUNCTION_REQUEST.format(request_uuid))
        s = config.aurorax.search.requests.get_status(url)
    except pyaurorax.AuroraXAPIError as e:
        if ("API error code 404" in str(e)):
            click.echo("%s occurred: request ID not found" % (type(e).__name__))
        else:  # pragma: nocover-ok
            click.echo("%s occurred: unexpected error -- %s" % (type(e).__name__, e.args[0]))
        sys.exit(1)
    except pyaurorax.AuroraXError as e:  # pragma: nocover-ok
        click.echo("%s occurred: %s" % (type(e).__name__, e.args[0]))
        sys.exit(1)

    # print out query
    if ("query" in s["search_request"]):
        query_to_show = s["search_request"]["query"]
        del query_to_show["request_id"]
        click.echo(pprint.pformat(query_to_show))
    else:
        click.echo("\nSearch query missing from request status, unable to display")  # pragma: nocover


@conjunctions_group.command("get_data", short_help="Get data for a conjunction search request")
@click.argument("request_uuid", type=str)
@click.option("--outfile", type=str, help="Output file to save data to (a .json file)")
@click.option("--output-to-terminal", type=click.Choice(["dict", "objects"]), help="Output data to terminal in a certain format (instead of to file)")
@click.option("--indent", type=int, default=2, show_default=True, help="Indentation when saving data to file")
@click.option("--minify", is_flag=True, help="Minify the JSON data saved to file")
@click.pass_obj
def get_data(config, request_uuid, outfile, output_to_terminal, indent, minify):
    """
    Get the data for a conjunction search request

    \b
    REQUEST_UUID    the request unique identifier
    """
    get_search_data(config.aurorax, "conjunctions", request_uuid, outfile, output_to_terminal, indent, minify)


@conjunctions_group.command("search_resubmit", short_help="Resubmit a conjunction search request")
@click.argument("request_uuid", type=str)
@click.pass_obj
def search_resubmit(config, request_uuid):
    """
    Resubmit a conjunction search request

    \b
    REQUEST_UUID    the request unique identifier
    """
    # get request status
    try:
        click.echo("Retrieving query for request '%s' ..." % (request_uuid))
        url = "%s/%s" % (config.aurorax.api_base_url, URL_SUFFIX_CONJUNCTION_REQUEST.format(request_uuid))
        status = config.aurorax.search.requests.get_status(url)
    except pyaurorax.AuroraXAPIError as e:
        if ("API error code 404" in str(e)):
            click.echo("%s occurred: request ID not found" % (type(e).__name__))
        else:  # pragma: nocover-ok
            click.echo("%s occurred: unexpected error -- %s" % (type(e).__name__, e.args[0]))
        sys.exit(1)
    except pyaurorax.AuroraXError as e:  # pragma: nocover-ok
        click.echo("%s occurred: %s" % (type(e).__name__, e.args[0]))
        sys.exit(1)

    # set the query to use for resubmission
    if ("query" not in status["search_request"]):  # pragma: nocover
        click.echo("Error resubmitting: missing query from original request ID")
        sys.exit(1)
    q = status["search_request"]["query"]

    # create search object
    click.echo("Preparing new search ...")
    s = __create_search_object_from_query(config.aurorax, q)

    # submit search
    click.echo("Submitting new search ...")
    s.execute()

    # output new request ID
    click.echo("Request has been resubmitted, new request ID is %s" % (s.request_id))


@conjunctions_group.command("search_template", short_help="Output template for a conjunction search request")
@click.option("--outfile", type=str, help="Save template to a file")
@click.option("--indent", type=int, default=2, show_default=True, help="Indentation to use when outputting template")
@click.pass_obj
def search_template(config, outfile, indent):
    """
    Output template for a conjunction search request
    """
    if (outfile is not None):
        with open(outfile, 'w', encoding="utf-8") as fp:
            json.dump(CONJUNCTION_SEARCH_TEMPLATE, fp, indent=indent)  # type: ignore
        click.echo("Saved template to %s" % (outfile))
    else:
        click.echo(json.dumps(CONJUNCTION_SEARCH_TEMPLATE, indent=indent))


@conjunctions_group.command("search", short_help="Perform a conjunction search request")
@click.argument("infile", type=str)
@click.option("--poll-interval", default=__STANDARD_POLLING_SLEEP_TIME, show_default=True, help="Polling interval when waiting for data (seconds)")
@click.option("--outfile", type=str, help="Output file to save data to (a .json file)")
@click.option("--output-to-terminal", type=click.Choice(["dict", "objects"]), help="Output data to terminal in a certain format (instead of to file)")
@click.option("--indent", type=int, default=2, show_default=True, help="indentation when saving data to file")
@click.option("--minify", is_flag=True, help="Minify the JSON data saved to file")
@click.option("--quiet", is_flag=True, help="Quiet output")
@click.option("--swarmaurora-show-url", is_flag=True, help="Display the URL for displaying conjunction results in Swarm-Aurora")
@click.option("--swarmaurora-open-browser", is_flag=True, help="In a browser, open the conjunction results in Swarm-Aurora")
@click.option("--swarmaurora-browser-type",
              default="default",
              type=click.Choice(__BROWSER_TYPES),
              help="Output data to terminal in a certain format (instead of to file)")
@click.option("--swarmaurora-save-custom-import-file", is_flag=True, help="Generate a Swarm-Aurora custom import file for the search")
@click.pass_obj
def search(config, infile, poll_interval, outfile, output_to_terminal, indent, minify, quiet, swarmaurora_show_url, swarmaurora_open_browser,
           swarmaurora_browser_type, swarmaurora_save_custom_import_file):
    """
    Perform a conjunction search request

    \b
    INFILE      input file with query (must be a JSON)
    """
    # check that infile exists
    if not (os.path.exists(infile)):
        click.echo("Error: infile doesn't exist (%s)" % (infile))
        sys.exit(1)

    # read in infile
    if (quiet is False):
        click.echo("[%s] Reading in query file ..." % (datetime.datetime.now()))
    with open(infile, 'r', encoding="utf-8") as fp:
        q = json.load(fp)

    # set search params
    if (quiet is False):
        click.echo("[%s] Preparing search ..." % (datetime.datetime.now()))
    start = parse(q["start"], ignoretz=True)
    end = parse(q["end"], ignoretz=True)
    max_distances = q["max_distances"]
    ground = [] if "ground" not in q else q["ground"]
    space = [] if "space" not in q else q["space"]
    events = [] if "events" not in q else q["events"]
    custom = [] if "custom" not in q else q["custom"]
    conjunction_types = [] if "conjunction_types" not in q else q["conjunction_types"]
    verbose_search = True if quiet is False else False

    # start search
    s = config.aurorax.search.conjunctions.search(start,
                                                  end,
                                                  distance=max_distances,
                                                  ground=ground,
                                                  space=space,
                                                  events=events,
                                                  custom_locations=custom,
                                                  conjunction_types=conjunction_types,
                                                  poll_interval=poll_interval,
                                                  verbose=verbose_search,
                                                  return_immediately=True)

    # wait for data
    s.wait(poll_interval=poll_interval, verbose=verbose_search)

    # search has finished, save results to a file or output to terminal
    get_search_data(config.aurorax, "conjunctions", s.request_id, outfile, output_to_terminal, indent, minify, show_times=True, search_obj=s)

    # do Swarm-Aurora tasks if requested
    if (swarmaurora_save_custom_import_file is True):
        print("[%s] Generating Swarm-Aurora custom import file ..." % (datetime.datetime.now()))
        swarmaurora_custom_import_filename = config.aurorax.search.conjunctions.swarmaurora.create_custom_import_file(s)
        print("[%s] Saved Swarm-Aurora custom import file to '%s'" % (datetime.datetime.now(), swarmaurora_custom_import_filename))
    if (swarmaurora_show_url is True):
        url = config.aurorax.search.conjunctions.swarmaurora.get_url(s)
        print("[%s] Swarm-Aurora URL: %s" % (datetime.datetime.now(), url))
    if (swarmaurora_open_browser is True):  # pragma: nocover-ok
        print(("[%s] Launching a browser to show this conjunction search in "
               "Swarm-Aurora (browser='%s') ...") % (datetime.datetime.now(), swarmaurora_browser_type))
        if (swarmaurora_browser_type == "default"):
            swarmaurora_browser_type = None
        config.aurorax.search.conjunctions.swarmaurora.open_in_browser(s, browser=swarmaurora_browser_type)


@conjunctions_group.command("describe", short_help="Describe a conjunction search request")
@click.argument("infile", type=str)
@click.pass_obj
def describe(config, infile):
    """
    Describe a conjunction search request using
    "SQL-like" syntax

    \b
    INFILE      input file with query (must be a JSON)
    """
    # check that infile exists
    if not (os.path.exists(infile)):
        click.echo("Error: infile doesn't exist (%s)" % (infile))
        sys.exit(1)

    # read in infile
    with open(infile, 'r', encoding="utf-8") as fp:
        q = json.load(fp)

    # create search object
    s = __create_search_object_from_query(config.aurorax, q)

    # describe the search
    d = config.aurorax.search.conjunctions.describe(s)

    # output
    click.echo(d)



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Data source statistics information
"""

import datetime
from typing import Optional


class DataSourceStatistics:
    """
    Data source statistics information

    Attributes:
        ephemeris_count (int): 
            Total number of ephemeris records for this data source

        data_product_count (int): 
            Total number of ephemeris records for this data source

        earliest_ephemeris_loaded (datetime.datetime): 
            Timestamp of the earliest ephemeris record

        latest_ephemeris_loaded (datetime.datetime): 
            Timestamp of the latest ephemeris record

        earliest_data_product_loaded (datetime.datetime): 
            Timestamp of the earliest data_product record
            
        latest_data_product_loaded (datetime.datetime): 
            Timestamp of the latest data product record
    """

    def __init__(self,
                 ephemeris_count: int,
                 data_product_count: int,
                 earliest_ephemeris_loaded: Optional[datetime.datetime] = None,
                 latest_ephemeris_loaded: Optional[datetime.datetime] = None,
                 earliest_data_product_loaded: Optional[datetime.datetime] = None,
                 latest_data_product_loaded: Optional[datetime.datetime] = None):
        self.ephemeris_count = ephemeris_count
        self.data_product_count = data_product_count
        self.earliest_ephemeris_loaded = earliest_ephemeris_loaded
        self.latest_ephemeris_loaded = latest_ephemeris_loaded
        self.earliest_data_product_loaded = earliest_data_product_loaded
        self.latest_data_product_loaded = latest_data_product_loaded

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return "DataSourceStatistics(ephemeris_count=%d, data_product_count=%d, ...)" % (
            self.ephemeris_count,
            self.data_product_count,
        )

    def pretty_print(self):
        """
        A special print output for this class.
        """
        print("DataSourceStatistics:")
        print("  %-30s: %d" % ("ephemeris_count", self.ephemeris_count))
        print("  %-30s: %d" % ("data_product_count", self.data_product_count))
        print("  %-30s: %s" % ("earliest_ephemeris_loaded", self.earliest_ephemeris_loaded))
        print("  %-30s: %s" % ("latest_ephemeris_loaded", self.latest_ephemeris_loaded))
        print("  %-30s: %s" % ("earliest_data_product_loaded", self.earliest_data_product_loaded))
        print("  %-30s: %s" % ("latest_data_product_loaded", self.latest_data_product_loaded))



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
from .classes.data_source import DataSource, FORMAT_FULL_RECORD
from .classes.data_source_stats import DataSourceStatistics
from ..api import AuroraXAPIRequest
from ...exceptions import (
    AuroraXNotFoundError,
    AuroraXDuplicateError,
    AuroraXConflictError,
    AuroraXAPIError,
)
from ..._util import show_warning


def __serialize_data_source_stats(stats):
    """
    Takes in raw JSON format of data source stats, serializes it 
    to a DataSourceStatistics object.
    
    NOTE: this is done specially instead of DataSourceStatistics(**ds) because
    the data source stats item returned by the API includes the data source itself, 
    and we don't need that extra bit. We strip it out when creating this
    DataSourceStatistics object.
    """
    # set ephemeris values
    earliest_ephemeris_loaded = None
    latest_ephemeris_loaded = None
    if (stats["earliest_ephemeris_loaded"] is not None):
        earliest_ephemeris_loaded = datetime.datetime.fromisoformat(stats["earliest_ephemeris_loaded"])
    if (stats["latest_ephemeris_loaded"] is not None):
        latest_ephemeris_loaded = datetime.datetime.fromisoformat(stats["latest_ephemeris_loaded"])

    # set data product values
    earliest_data_product_loaded = None
    latest_data_product_loaded = None
    if (stats["earliest_data_product_loaded"] is not None):
        earliest_data_product_loaded = datetime.datetime.fromisoformat(stats["earliest_data_product_loaded"])
    if (stats["latest_data_product_loaded"] is not None):
        latest_data_product_loaded = datetime.datetime.fromisoformat(stats["latest_data_product_loaded"])

    # return
    return DataSourceStatistics(
        ephemeris_count=stats["ephemeris_count"],
        data_product_count=stats["data_product_count"],
        earliest_ephemeris_loaded=earliest_ephemeris_loaded,
        latest_ephemeris_loaded=latest_ephemeris_loaded,
        earliest_data_product_loaded=earliest_data_product_loaded,
        latest_data_product_loaded=latest_data_product_loaded,
    )


def list(aurorax_obj, program, platform, instrument_type, source_type, owner, format, order, include_stats):
    # make request
    params = {
        "program": program,
        "platform": platform,
        "instrument_type": instrument_type,
        "source_type": source_type,
        "owner": owner,
        "format": format,
        "include_stats": include_stats,
    }
    url = "%s/%s" % (aurorax_obj.api_base_url, aurorax_obj.search.api.URL_SUFFIX_DATA_SOURCES)
    req = aurorax_obj.search.api.AuroraXAPIRequest(aurorax_obj, method="get", url=url, params=params)
    res = req.execute()

    # order results
    res.data = sorted(res.data, key=lambda x: "None" if x[order] is None else x[order])

    # cast
    sources = []
    if len(res.data):
        for ds in res.data:
            source = DataSource(**ds, format=format)
            if ("stats" in ds and ds["stats"] is not None):
                source.stats = __serialize_data_source_stats(ds["stats"])
            sources.append(source)
    else:
        return []

    # remove not_applicable sources
    sources_pruned = []
    for source in sources:
        if (source.identifier is not None and source.identifier >= 0):
            # exclude under-the-hood adhoc data sources
            sources_pruned.append(source)

    # return
    return sources_pruned


def search(aurorax_obj, programs, platforms, instrument_types, format, order, include_stats):
    # make request
    request_data = {
        "programs": programs,
        "platforms": platforms,
        "instrument_types": instrument_types,
    }
    params = {
        "format": format,
        "include_stats": include_stats,
    }
    url = "%s/%s" % (aurorax_obj.api_base_url, aurorax_obj.search.api.URL_SUFFIX_DATA_SOURCES_SEARCH)
    req = aurorax_obj.search.api.AuroraXAPIRequest(aurorax_obj, method="post", url=url, params=params, body=request_data)
    res = req.execute()

    # order results
    res.data = sorted(res.data, key=lambda x: "None" if x[order] is None else x[order])

    # cast
    sources = []
    if len(res.data):
        for ds in res.data:
            source = DataSource(**ds, format=format)
            if ("stats" in ds and ds["stats"] is not None):
                source.stats = __serialize_data_source_stats(ds["stats"])
            sources.append(source)
    else:
        return []

    # return
    return sources


def get(aurorax_obj, program, platform, instrument_type, format, include_stats):
    # get the data source
    sources = list(aurorax_obj=aurorax_obj,
                   program=program,
                   platform=platform,
                   instrument_type=instrument_type,
                   source_type=None,
                   owner=None,
                   format=format,
                   order="identifier",
                   include_stats=include_stats)

    # set results to the first thing found
    if (len(sources) == 1):
        return sources[0]
    elif (len(sources) > 1):  # pragma: nocover-ok
        show_warning("Found more than one data source matching this criteria, returning the first (found %d)" % (len(sources)), stacklevel=1)
        return sources[0]
    else:
        raise AuroraXNotFoundError("No matching data source found")


def get_using_filters(aurorax_obj, program, platform, instrument_type, source_type, owner, format, order, include_stats):
    # get data sources
    sources = list(aurorax_obj=aurorax_obj,
                   program=program,
                   platform=platform,
                   instrument_type=instrument_type,
                   source_type=source_type,
                   owner=owner,
                   format=format,
                   order=order,
                   include_stats=include_stats)

    # return
    return sources


def get_using_identifier(aurorax_obj, identifier, format, include_stats):
    # make request
    params = {
        "format": format,
        "include_stats": include_stats,
    }
    url = "%s/%s/%d" % (aurorax_obj.api_base_url, aurorax_obj.search.api.URL_SUFFIX_DATA_SOURCES, identifier)
    req = AuroraXAPIRequest(aurorax_obj, method="get", url=url, params=params)
    res = req.execute()

    # cast
    data_source = DataSource(**res.data, format=format)
    if ("stats" in res.data and res.data["stats"] is not None):
        data_source.stats = __serialize_data_source_stats(res.data["stats"])

    # return
    return data_source


def add(aurorax_obj, ds):
    # check that the length of the program, platform, and instrument type is
    # within API constraints
    #
    # NOTE: this check is done here because the API will return a generic 500
    # error with no details that the issue was due to these values being too
    # long. If we run into situation where these limits need to be increased,
    # we can definitely do it. We just need to change the below IF-statements,
    # and change the API's database code. The constraints are embedded in the API
    # at the database level (within the data sources table constraints). Also
    # note that the source type also has a limit, but because we control these
    # ourselves, we don't need a check for it.
    if not all([ds.program, ds.platform, ds.instrument_type, ds.source_type, ds.display_name]):
        raise ValueError("Missing required fields. To create a data source, the program, platform, " +
                         "instrument_type, source_type, and display_name, are all required")
    if (len(ds.program) > 200):
        raise ValueError("Program too long. Must be less than 200 characters")
    if (len(ds.platform) > 50):
        raise ValueError("Platform too long. Must be less than 50 characters")
    if (len(ds.instrument_type) > 200):
        raise ValueError("Instrument type too long. Must be less than 200 characters")
    if (len(ds.display_name) > 50):
        raise ValueError("Display name too long. Must be less than 50 characters")

    # set up request
    request_data = {
        "program": ds.program,
        "platform": ds.platform,
        "instrument_type": ds.instrument_type,
        "source_type": ds.source_type,
        "display_name": ds.display_name,
        "ephemeris_metadata_schema": ds.ephemeris_metadata_schema,
        "data_product_metadata_schema": ds.data_product_metadata_schema,
        "metadata": ds.metadata
    }
    if (ds.identifier is not None):
        request_data["identifier"] = ds.identifier

    # make request
    url = "%s/%s" % (aurorax_obj.api_base_url, aurorax_obj.search.api.URL_SUFFIX_DATA_SOURCES)
    req = AuroraXAPIRequest(aurorax_obj, method="post", url=url, body=request_data)
    res = req.execute()

    # evaluate response
    if (res.status_code == 409):
        raise AuroraXDuplicateError("%s - %s" % (res.data["error_code"], res.data["error_message"]))

    # return
    try:
        return DataSource(**res.data)
    except Exception:  # pragma: nocover
        raise AuroraXAPIError("Could not create data source") from None


def delete(aurorax_obj, identifier):
    # do request
    url = "%s/%s/%d" % (aurorax_obj.api_base_url, aurorax_obj.search.api.URL_SUFFIX_DATA_SOURCES, identifier)
    req = AuroraXAPIRequest(aurorax_obj, method="delete", url=url, null_response=True)
    res = req.execute()

    # evaluate response
    if (res.status_code == 400):  # pragma: nocover
        raise AuroraXAPIError("%s - %s" % (res.data["error_code"], res.data["error_message"]))
    elif (res.status_code == 409):
        raise AuroraXConflictError("%s - %s" % (res.data["error_code"], res.data["error_message"]))

    # return
    return 0


def update(aurorax_obj, identifier, program, platform, instrument_type, source_type, display_name, metadata, owner, maintainers,
           ephemeris_metadata_schema, data_product_metadata_schema):
    # create a data source
    ds = DataSource(identifier=identifier,
                    program=program,
                    platform=platform,
                    instrument_type=instrument_type,
                    source_type=source_type,
                    display_name=display_name,
                    metadata=metadata,
                    owner=owner,
                    maintainers=maintainers,
                    ephemeris_metadata_schema=ephemeris_metadata_schema,
                    data_product_metadata_schema=data_product_metadata_schema)

    # set URL and request data
    url = "%s/%s/%d" % (aurorax_obj.api_base_url, aurorax_obj.search.api.URL_SUFFIX_DATA_SOURCES, ds.identifier)
    req_data = ds.__dict__
    del req_data["stats"]
    del req_data["format"]

    # make request
    req = AuroraXAPIRequest(aurorax_obj, method="patch", url=url, body=req_data)
    req.execute()

    # return
    try:
        return get_using_identifier(aurorax_obj, ds.identifier, FORMAT_FULL_RECORD, False)
    except Exception:  # pragma: nocover
        raise AuroraXAPIError("Could not update data source") from None



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Class definition for a conjunction
"""

import datetime
from typing import Dict, List
from ...sources import DataSource

# conjunction type - north b-trace
CONJUNCTION_TYPE_NBTRACE: str = "nbtrace"
"""
Conjunction search 'conjunction_type' category for finding conjunctions using the north B-trace data
"""

# conjunction type - south b-trace
CONJUNCTION_TYPE_SBTRACE: str = "sbtrace"
"""
Conjunction search 'conjunction_type' category for finding conjunctions using the south B-trace data
"""

# conjunction type - geographic
CONJUNCTION_TYPE_GEOGRAPHIC: str = "geographic"
"""
Conjunction search 'conjunction_type' category for finding conjunctions using the geographic position data
"""


class Conjunction:
    """
    Conjunction object

    Attributes:
        conjunction_type (str): 
            The type of location data used when the conjunction was found (either `nbtrace`, 
            `sbtrace`, or `geographic`)
            
        start (datetime.datetime): 
            Start timestamp of the conjunction

        end (datetime.datetime): 
            End timestamp of the conjunction

        data_sources (List[DataSource]): 
            Data sources in the conjunction

        min_distance (float): 
            Minimum kilometer distance of the conjunction

        max_distance (float): 
            Maximum kilometer distance of the conjunction

        events (List[Dict]): 
            The sub-conjunctions that make up this over-arching conjunction (the conjunctions 
            between each set of two data sources)

        closest_epoch (datetime.datetime): 
            Timestamp for when data sources were closest

        farthest_epoch (datetime.datetime): 
            Timestamp for when data sources were farthest
    """

    def __init__(
        self,
        conjunction_type: str,
        start: datetime.datetime,
        end: datetime.datetime,
        data_sources: List[DataSource],
        min_distance: float,
        max_distance: float,
        events: List[Dict],
        closest_epoch: datetime.datetime,
        farthest_epoch: datetime.datetime,
    ):
        self.conjunction_type = conjunction_type
        self.start = start
        self.end = end
        self.data_sources = data_sources
        self.min_distance = min_distance
        self.max_distance = max_distance
        self.events = events
        self.closest_epoch = closest_epoch
        self.farthest_epoch = farthest_epoch

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return f"Conjunction(start={repr(self.start)}, end={repr(self.end)}, min_distance={self.min_distance:.2f}, " \
            f"max_distance={self.max_distance:.2f}, data_sources=[...], events=[...])"

    def pretty_print(self):
        """
        A special print output for this class.
        """
        # set special strings
        if (len(self.data_sources) == 1):  # pragma: nocover-ok
            data_sources_str = "[%d data source]" % (len(self.data_sources))
        else:
            data_sources_str = "[%d data sources]" % (len(self.data_sources))
        if (len(self.events) == 1):
            events_str = "[%d event]" % (len(self.events))
        else:
            events_str = "[%d events]" % (len(self.events))

        # print
        print("Conjunction:")
        print("  %-18s: %s" % ("conjunction_type", self.conjunction_type))
        print("  %-18s: %s" % ("start", self.start))
        print("  %-18s: %s" % ("end", self.end))
        print("  %-18s: %s" % ("data_sources", data_sources_str))
        print("  %-18s: %s" % ("min_distance", self.min_distance))
        print("  %-18s: %s" % ("max_distance", self.max_distance))
        print("  %-18s: %s" % ("events", events_str))
        print("  %-18s: %s" % ("closest_epoch", self.closest_epoch))
        print("  %-18s: %s" % ("farthest_epoch", self.farthest_epoch))



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Methods for working with data in a specific bounding box.
"""

# imports for this file
from .extract_metric import ExtractMetricManager

__all__ = ["BoundingBoxManager"]


class BoundingBoxManager:
    """
    The BoundingBoxManager object is initialized within every PyAuroraX object. It acts as a way to access 
    the submodules and carry over configuration information in the super class.
    """

    def __init__(self, aurorax_obj):
        # initialize super class object
        self.__aurorax_obj = aurorax_obj

        # initialize sub-modules
        self.__extract_metric = ExtractMetricManager(self.__aurorax_obj)

    # ------------------------------------------
    # properties for submodule managers
    # ------------------------------------------
    @property
    def extract_metric(self):
        """
        Access to the `extract_metric` submodule from within a PyAuroraX object.
        """
        return self.__extract_metric



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Use the AuroraX search engine to find conjunctions between groupings 
of data sources.

Note that all functions and classes from submodules are all imported
at this level of the conjunctions module. They can be referenced from
here instead of digging in deeper to the submodules.
"""

import datetime
from typing import Dict, Optional, Union, Sequence, Literal
from .swarmaurora import SwarmAuroraManager
from .classes.search import ConjunctionSearch
from .classes.criteria_block import (
    GroundCriteriaBlock,
    SpaceCriteriaBlock,
    EventsCriteriaBlock,
    CustomLocationsCriteriaBlock,
)
from ._conjunctions import search as func_search
from ._conjunctions import search_from_raw_query as func_search_from_raw_query
from ._conjunctions import describe as func_describe
from ._conjunctions import get_request_url as func_get_request_url
from ._conjunctions import create_advanced_distance_combos as func_create_advanced_distance_combos

__all__ = ["ConjunctionsManager"]


class ConjunctionsManager:
    """
    The ConjunctionsManager object is initialized within every PyAuroraX object. It acts as a way to access 
    the submodules and carry over configuration information in the super class.
    """

    __STANDARD_POLLING_SLEEP_TIME: float = 1.0  # Polling sleep time when waiting for data (after the initial sleep time), in seconds

    def __init__(self, aurorax_obj):
        self.__aurorax_obj = aurorax_obj

        # initialize sub-modules
        self.__swarmaurora = SwarmAuroraManager(self.__aurorax_obj)

    @property
    def swarmaurora(self):
        """
        Access to the `swarmaurora` submodule from within a PyAuroraX object.
        """
        return self.__swarmaurora

    def search(self,
               start: datetime.datetime,
               end: datetime.datetime,
               distance: Union[int, float, Dict],
               ground: Sequence[Union[GroundCriteriaBlock, Dict]] = [],
               space: Sequence[Union[SpaceCriteriaBlock, Dict]] = [],
               events: Sequence[Union[EventsCriteriaBlock, Dict]] = [],
               custom_locations: Sequence[Union[CustomLocationsCriteriaBlock, Dict]] = [],
               conjunction_types: Sequence[Union[str, Literal["nbtrace", "sbtrace", "geographic"]]] = [],
               response_format: Optional[Dict] = None,
               poll_interval: float = __STANDARD_POLLING_SLEEP_TIME,
               return_immediately: bool = False,
               verbose: bool = False) -> ConjunctionSearch:
        """
        Search for conjunctions

        By default, this function will block and wait until the request completes and
        all data is downloaded. If you don't want to wait, set the 'return_immediately`
        value to True. The Search object will be returned right after the search has been
        started, and you can use the helper functions as part of that object to get the
        data when it's done.

        Args:
            start (datetime.datetime): 
                Start timestamp of the search (inclusive).

            end (datetime.datetime): 
                End timestamp of the search (inclusive).

            distance (int or float or Dict): 
                The maximum distance allowed between data sources when searching for
                conjunctions. This can either be a number (int or float), or a dictionary
                modified from the output of the "get_advanced_distances_combos()" function.

            ground (List[GroundCriteriaBlock or Dict]): 
                List of ground instrument criteria blocks, defaults to [].

            space (List[SpaceCriteriaBlock or Dict]): 
                List of space instrument criteria blocks, defaults to [].

            events (List[EventsCriteriaBlock or Dict]): 
                List of event criteria blocks, defaults to [].

            custom_locations (List[CustomLocationsCriteriaBlock or Dict]): 
                List of custom location criteria blocks, defaults to [].

            conjunction_types (List[str]): 
                List of conjunction types, defaults to [] (meaning all conjunction types). Valid
                options are 'nbtrace', 'sbtrace', and 'geographic'. Defaults to 'nbtrace'.

            response_format (Dict): 
                JSON representation of desired data response format.

            poll_interval (bool): 
                Seconds to wait between polling calls, defaults to 1 second.

            return_immediately (bool): 
                Initiate the search and return without waiting for data to be received, defaults 
                to `False`.

            verbose (bool): 
                Show the progress of the request using the request log, defaults to `False`.

        Returns:
            A `pyaurorax.search.ConjunctionSearch` object

        Raises:
            pyaurorax.exceptions.AuroraXSearchError: The API experienced a search error
        """
        return func_search(
            self.__aurorax_obj,
            start,
            end,
            distance,
            ground,
            space,
            events,
            custom_locations,
            conjunction_types,
            response_format,
            poll_interval,
            return_immediately,
            verbose,
        )

    def search_from_raw_query(self,
                              query: Union[Dict, str],
                              poll_interval: float = __STANDARD_POLLING_SLEEP_TIME,
                              return_immediately: bool = False,
                              verbose: bool = False) -> ConjunctionSearch:
        """
        Search for conjunctions, using a query dictionary as the input. 
        
        This is especially useful if you're working in the AuroraX Conjunction Search webpage 
        to create a search and you'd like to port it over to this Python library with ease.

        Args:
            query (Dict or str): 
                A query in dictionary or string format. If it's in string format, it should valid
                JSON.

                In the conjunction search web page (https://aurorax.space/conjunctionSearch/standard), 
                click on the 'More' button under the 'Tools' section on the right of the page. Then click 
                on 'About query' to bring up a modal. Copy the query in JSON format using the clipboard 
                icon, and use this JSON string as input to the function.

                Many JSON strings can be pasted as a dictionary object without any adjustments, however, 
                there are a few edge cases. For example, if the max_distances field has 'null' values in 
                it, then it is not valid Python. In this case, enclose the JSON in a multi-line string 
                (using the triple-quotes), and pass the query as a string to the function.

            poll_interval (bool): 
                Seconds to wait between polling calls, defaults to 1 second.

            return_immediately (bool): 
                Initiate the search and return without waiting for data to be received, defaults 
                to `False`.

            verbose (bool): 
                Show the progress of the request using the request log, defaults to `False`.

        Returns:
            A `ConjunctionSearch` object.

        Raises:
            pyaurorax.exceptions.AuroraXSearchError: An error was encountered during the search process
        """
        return func_search_from_raw_query(self.__aurorax_obj, query, poll_interval, return_immediately, verbose)

    def describe(self, search_obj: Optional[ConjunctionSearch] = None, query_dict: Optional[Dict] = None) -> str:
        """
        Describe a conjunction search as an "SQL-like" string. Either a ConjunctionSearch
        object can be supplied, or a dictionary of the raw JSON query.

        Args:
            search_obj (ConjunctionSearch): 
                The conjunction search to describe, optional

            query_dict (Dict): 
                The conjunction search query represented as a raw dictionary, optional

        Returns:
            The "SQL-like" string describing the conjunction search object

        Raises:
            pyaurorax.exceptions.AuroraXError: Invalid arguments
        """
        return func_describe(self.__aurorax_obj, search_obj, query_dict)

    def get_request_url(self, request_id: str) -> str:
        """
        Get the conjunction search request URL for a given request ID. This URL 
        can be used for subsequent pyaurorax.requests function calls. Primarily 
        this method facilitates delving into details about a set of already-submitted 
        conjunction searches.

        Args:
            request_id (str): 
                The request identifier

        Returns:
            The request URL
        """
        return func_get_request_url(self.__aurorax_obj, request_id)

    def create_advanced_distance_combos(
        self,
        distance: Optional[int] = None,
        ground: int = 0,
        space: int = 0,
        events: int = 0,
        custom: int = 0,
    ) -> Dict:
        """
        Get the advanced distances combinations for the specified parameters

        Args:
            distance (int): 
                The default distance to use, defaults to None
            
            ground (int): 
                The number of ground criteria blocks, defaults to 0

            space (int): 
                The number of space criteria blocks, defaults to 0

            events (int): 
                The number of events criteria blocks, defaults to 0

            custom (int): 
                The number of custom location criteria blocks, defaults to 0

        Returns:
            The advanced distances combinations
        """
        return func_create_advanced_distance_combos(distance, ground, space, events, custom)

    def create_response_format_template(self, default: bool = True) -> Dict:
        """
        Generate a template dictionary that can be used as the response_format parameter
        in a conjunction search.

        Args:
            default (bool): 
                The default value to set for every parameter that can be returned, defaults
                to True.

        Returns:
            A template dictionary for the response format
        """
        return {
            "conjunction_type": default,
            "start": default,
            "end": default,
            "min_distance": default,
            "max_distance": default,
            "closest_epoch": default,
            "farthest_epoch": default,
            "data_sources": {
                "identifier": default,
                "program": default,
                "platform": default,
                "instrument_type": default,
                "source_type": default,
                "display_name": default,
                "ephemeris_metadata_schema": {
                    "field_name": default,
                    "description": default,
                    "data_type": default,
                    "allowed_values": default,
                    "additional_description": default
                },
                "data_product_metadata_schema": {
                    "field_name": default,
                    "description": default,
                    "data_type": default,
                    "allowed_values": default,
                    "additional_description": default
                },
                "owner": default,
                "maintainers": default,
                "metadata": default
            },
            "events": {
                "conjunction_type": default,
                "e1_source": {
                    "identifier": default,
                    "program": default,
                    "platform": default,
                    "instrument_type": default,
                    "source_type": default,
                    "display_name": default,
                    "ephemeris_metadata_schema": {
                        "field_name": default,
                        "description": default,
                        "data_type": default,
                        "allowed_values": default,
                        "additional_description": default
                    },
                    "data_product_metadata_schema": {
                        "field_name": default,
                        "description": default,
                        "data_type": default,
                        "allowed_values": default,
                        "additional_description": default
                    },
                    "owner": default,
                    "maintainers": default,
                    "metadata": default
                },
                "e2_source": {
                    "identifier": default,
                    "program": default,
                    "platform": default,
                    "instrument_type": default,
                    "source_type": default,
                    "display_name": default,
                    "ephemeris_metadata_schema": {
                        "field_name": default,
                        "description": default,
                        "data_type": default,
                        "allowed_values": default,
                        "additional_description": default
                    },
                    "data_product_metadata_schema": {
                        "field_name": default,
                        "description": default,
                        "data_type": default,
                        "allowed_values": default,
                        "additional_description": default
                    },
                    "owner": default,
                    "maintainers": default,
                    "metadata": default
                },
                "start": default,
                "end": default,
                "min_distance": default,
                "max_distance": default,
                "generated_e1_ephemeris_query": {
                    "request_id": default,
                    "data_sources": {
                        "programs": default,
                        "platforms": default,
                        "instrument_types": default,
                        "ephemeris_metadata_filters": {
                            "logicalOperator": default,
                            "expressions": {
                                "key": default,
                                "operator": default,
                                "values": default
                            }
                        }
                    },
                    "start": default,
                    "end": default
                },
                "generated_e2_ephemeris_query": {
                    "request_id": default,
                    "data_sources": {
                        "programs": default,
                        "platforms": default,
                        "instrument_types": default,
                        "ephemeris_metadata_filters": {
                            "logicalOperator": default,
                            "expressions": {
                                "key": default,
                                "operator": default,
                                "values": default
                            }
                        }
                    },
                    "start": default,
                    "end": default
                }
            }
        }



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Utility methods. For example, converting arbitrary geographic locations to North/South 
B-trace geographic locations.
"""

import datetime
from ..location import Location
from ._calculate_btrace import ground_geo_to_nbtrace as func_ground_geo_to_nbtrace
from ._calculate_btrace import ground_geo_to_sbtrace as func_ground_geo_to_sbtrace


class UtilManager:
    """
    The UtilManager object is initialized within every PyAuroraX object. It acts as a way to access 
    the submodules and carry over configuration information in the super class.
    """

    def __init__(self):
        pass

    def ground_geo_to_nbtrace(self, geo_location: Location, timestamp: datetime.datetime) -> Location:
        """
        Convert geographic location to North B-Trace geographic location

        The timestamp is required because when calculating the B-trace values, the 
        location is converted into geomagnetic coordinates using AACGM. This conversion 
        is different  based on the timestamp since the magnetic coordinates change over time.

        Args:
            geo_location (Location): 
                A Location object representing the geographic location

            timestamp (datetime.datetime): 
                Timestamp for this set of latitudes and longitudes

        Returns:
            The north B-trace location as a `Location` object
        """
        return func_ground_geo_to_nbtrace(geo_location, timestamp)

    def ground_geo_to_sbtrace(self, geo_location: Location, timestamp: datetime.datetime) -> Location:
        """
        Convert geographic location to South B-Trace geographic location

        The timestamp is required because when calculating the B-trace values, the 
        location is converted into geomagnetic coordinates using AACGM. This conversion 
        is different  based on the timestamp since the magnetic coordinates change over time.

        Args:
            geo_location (Location): 
                A Location object representing the geographic location

            timestamp (datetime.datetime): 
                Timestamp for this set of latitudes and longitudes

        Returns:
            The south B-trace location as a `Location` object
        """
        return func_ground_geo_to_sbtrace(geo_location, timestamp)
